import numpy as np
from .error import TensorSyntaxError

def delete_blanks(element):
    
    return element.replace(" ","")

def correct_multiplication(element):
    
    s = 0

    while s < len(element)-1: 
 
        if (element[s] == ']' or element[s] == ')')  and (element[s+1].isalpha() == True or element[s+1] == '('):
        
            element = element[:s+1] + '*' + element[s+1:]  

        s = s+1
        
    return element

def parentesis_distribute(element):
    
    arr_paren = np.empty(len(element))   # arreglo de parentesis con sus indicadores
    arr_paren[:] = np.nan                # se inicializa como si no hubiese ningun parentesis
    
    arr_pareado = np.empty(len(element))   # arreglo de parentesis con sus indicadores
    arr_pareado[:] = np.nan                # se inicializa como si no hubiese ningun parentesis
    
    j = 0                                # contador de parentesis abiertos
    
    
    cant_par = 0

    
    for i in range(len(element)):
        
        if element[i] == '(':
                
            arr_paren[i] = j
            arr_pareado[i] = 1         # hay que parearlo con algun ')'
            
            cant_par = j+1             # le sumamos 1 porque el j empieza desde el 0. Aqui se cuentan la cantidad total de pares de parentesis
            
            j += 1

        elif element[i] == ')':
                
            l = i-1
                
            while np.isnan(arr_pareado[l]):    #  cuando se acaba el while, encuentra un '(' para ser pareado
                
                l = l-1
                
        
            arr_pareado[l] = np.nan
        
            arr_paren[i] = arr_paren[l]

    
    # entrega parentesis pareados indicados en el arr_paren

    return arr_paren,cant_par

def check_parentesis(arr_paren,element):
    
    for i in range(len(arr_paren)):
        
        if i > 0:
 
            if element[i] == '(' and (element[i-1] == 'D' or element[i-1] == 'C'):

                for j in range(i+1,len(arr_paren)):

                    if arr_paren[j] == arr_paren[i]:
                        
                        break
                        
                change = True
                
                for k in range(i,j-1):
                    
                    if element[k] == '+' or element[k] == '-':
                        
                        change = False
                        
                        break
                    
                if change == True:
                        
                    arr_paren[j] = arr_paren[i]
                        
                    arr_paren[i] = arr_paren[i]
                    
                    #sirve para DERIVADAS Y PARA COVARIANT DERIVATIVES

    # hasta aca el arrelo arr_paren solo tiene numeros en los parentesis que hay que expandir, es decir
    # se han eliminado todos los parentesis donde las derivadas aplican son productos, y sobre solamenete un tensor o de tipo
    # D(D(D(etc... ) ) )
    
    
    return arr_paren,element

def factores(i,j,element):
    
    list_factores = []
    
    factor = ''
    
    k = i+1

    while k <= j:
        
        if factor == '':
            
            factor +=  str(element[k])
                
        elif element[k] != '-' and element[k] != '+' and element[k] !=')':
            
            factor +=  str(element[k])
            
        elif element[k] == '-' or element[k] == '+':
            
            list_factores.append(factor)
            
            factor = str(element[k])

        elif element[k] == ')':
            
            list_factores.append(factor)
        
        k += 1
        
    return list_factores

#REVISAR ESTA FUNCION Y VER DONDE SE MULTIPLICAN Y REEMPLAZAN LOS FACTORES PARA QUE QUEDEN EN ELEMENT CUANDO 
#SE RETORNEN

def factores_der(i,j,element):
    

    element = element
    
    list_factores = []
    
    factor = ''
    
    k = i+1

    while k <= j:
   
        if factor == '':
            
            factor +=  str(element[k])
                
        elif element[k] != '-' and element[k] != '+':
            
            factor +=  str(element[k])
            
        elif element[k] == '-' or element[k] == '+':
            
            list_factores.append(factor)
            
            factor = str(element[k])

        if k == j:
            
            while factor[-1] == ',':
                
                factor = factor[:-1]
            
            list_factores.append(factor)
        
        k += 1
        
    return list_factores

#REVISAR ESTA FUNCION Y VER DONDE SE MULTIPLICAN Y REEMPLAZAN LOS FACTORES PARA QUE QUEDEN EN ELEMENT CUANDO 
#SE RETORNEN

def lado_der(i,j,element,arr_paren):
    
    factor = ''
    
    k = j+1
    
    while k < len(element):
        
        if element[k] == '(':    

            for l in range(k+1,len(arr_paren)):

                if arr_paren[l] == arr_paren[k]:
                        
                    break
                        
            factor += element[k:l+1]
            
            k = l
        
        
        
        elif (element[k-1:k+1] == '),' or element[k-1:k+1] == '],'):
            
            factor_der = factor
            
            break
            
        elif element[k] != '-' and element[k] != '+'  and  element[k] != ')' :
            
            factor +=  str(element[k])
            
        elif element[k] == '+' or element[k] == '-'  or element[k] == ')' :
            
            
            factor_der = factor    

            break
            
        if k == len(element)-1:
            
            factor_der = factor
        
        k += 1
        
    return k-1,factor_der

def lado_izq(i,j,element, arr_paren):
    
    factor = ''
    
    k = i-1 
    
    while k >= 0:
        
        if element[k] == '(':    

            for l in range(0,k+1):

                if arr_paren[l] == arr_paren[k]:
                        
                    break

            k = l

        if element[k] == '+' or element[k] == '-'  or  element[k] == '(' or k == 0:

            break
        
        k -= 1
    
    factor_izq = element[k:i]
    
    if factor_izq[0] == '(':
        
        factor_izq = factor_izq[1:]
    
    return k,factor_izq

def existe_algo(num,element):  #evalua si existe esa posicion de element
    
    if num < 0 or num >= len(element):
        
        bol = False
    
    else:
        
        bol = True
    
    return bol    

def evaluar(num,element):
    
    try:
        
        val = element[num] == '*'
    
    except:
        
        val = False
    
    return val

def evaluar_signo(num,element):
    
    try:
        
        if element[num] == '-' or element[num] == '+':

            val = True

        else:

            val = False
            
    except:
        
        val = False
    
    return val

def isDorCD(num,element):
    
    #retorne True or False
    #si es vdd q son D retorne 0 si es CD retorne 1
    
    try:
        
        if element[num] == 'D':

            bol = True
            val = 0
                
        elif element[num] == 'C':
            
            bol = True
            val = 1

        else:
            bol = False

            val = False
            
    except:
        
        bol = False
        val = False
    
    return bol,val  #bol dice si hay o no una derivada de cualquier tipo #val 0 Der, val 1 DerCov

def evaluar_paren_izq(num,element):
    
    try:
        
        if element[num] == '(':
            
            val = True
        
        else:
            
            val = False
    
    except:
        
        val = False
    
    return val

def evaluar_paren_der(num,element):
    
    try:
        
        if element[num] == ')' or element[num] == ',':
            
            val = True
        
        else:
            
            val = False
    
    except:
        
        val = False
    
    return val

def multiplicar(i,j,element,arr_paren):

    '''
    i posision de (
    j posision de )
    '''
    

    new_string = ''

    #print('Elemento en Multiplicar',element[i:j])

    if element[j-3] == ',' and isDorCD(i-1,element)[0] == False:  # aqui habra problema por el element[j-3] == ','

        #print(element[i-1:j-3])

        raise(TensorSyntaxError("Bad definition of derivative.\n Remember that it must be defined with 'C' or 'D', and check the '_' next to the derivative index."))
    
    
    if evaluar_paren_izq(i-1,element) and evaluar_paren_der(j+1,element):
        
        #HAY Q INCLUIR EL CASO DE Q ESTE PARENTESIS INTERNO RECIBIDO ESTE COMO: (()) O COMO ((),
        
        element = element[:i]  + element[i+1:j] +  element[j+1:]
    
    elif isDorCD(i-1,element)[0]:   #si es una derivada D o CD
        
        if isDorCD(i-1,element)[1] == 0:  #es una D
            
            cosa = 'D'
            
        else:                            #es una CD
            
            cosa = 'C'

        l = j
        
        while element[l] != '_':

            l = l - 1

        

        l = l - 3  #no entiendo esto


        #while element[l] == '_' and element[l-1] == ',':
        #    
        #    print('Bad definition of derivative')
        #    break
        #    #l = l - 3
        #    #D(T[_x],_x,_d,_s)

    # l+3 tiene la posicion donde esta la coma, es decir> entre (l+3)+1 <= cosas <= j-1 estan los elementos a derivar  
        
        cosa2 = element[l+2:j]
 
        list_factores = factores_der(i,l+2,element)
    
        for factor in list_factores:

            if factor[0] == '+':
                
                signo = '+'
                
                factor = factor[1:]
            
            elif factor[0] == '-':
                
                signo = '-'
                
                factor = factor[1:]
            
            else:
                
                signo = '+'
                 
            new_string += signo + cosa + '{' + factor + cosa2 + '}'
            
        try:

            if element[i-2] == '-':
                
                new_string = new_string.replace('+','BLABLASIGNO')
                new_string = new_string.replace('-','+')
                new_string = new_string.replace('BLABLASIGNO','-')
                
                element = element[:i-2] + '+' + '(' +new_string +')' + element[j+1:]
            
            elif element[i-2] == '*':
                
                element = element[:i-1]  + '(' +new_string +')' + element[j+1:]
                
            else:
                
                element = element[:i-1] + '+' + '(' +new_string +')' + element[j+1:]
        
        except:
        
            element = element[:i-1] +'(' +new_string +')' + element[j+1:]
    
    elif evaluar_signo(i-1,element) and evaluar(j+1,element) == False:
        #SI SOLO HAY SIGNO POR IZQUIERDA, ENTONCES MULTIPLICAR EL SIGNO (Y POR DERECHA NO HAY NADA)
        
        list_factores = factores(i,j,element)
        
        if element[i-1] == '-':
            
            new_string = ''
            
            for factor in list_factores:
                
                if factor[0] == '-':
                    
                    factor = '+' + factor[1:]
                
                elif factor[0] == '+':
                    
                    factor = '-' + factor[1:]
                    #CAMBIA LOS SIGNOS DE TODOS LSO FACTORES
                
                else:
                    
                    factor = '-' + factor
                    
                new_string += factor
                
            element = element[:i-1]+new_string+element[j+1:]
        
        else:
            #SI ES UN SIGNO + ENTONCES SOLO ELIMINE LOS PARENTESIS PORQ ES MULTPLICAR POR UN +
            
            element = element[:i-1]+element[i+1:j]+element[j+1:]
        
   #     element = element[1:-1]

    elif existe_algo(i-1,element) == False and evaluar(j+1,element) == False:
        #SI NO EXISTE NADA A LA IZQUERDA Y A LA DERECHA NO HAY MULTIPLICACION
        
        element = element[i+1:j]+element[j+1:]

    elif evaluar(i-1,element):

        pos_inicial, factor_izq = lado_izq(i,j,element, arr_paren)
        
        list_factores = factores(i,j,element)
            
        if factor_izq[0] == '+':
                    
            signo_factor_izq = 1
                
            factor_izq = factor_izq[1:]
                
        elif factor_izq[0] == '-':
                    
            signo_factor_izq = -1
                
            factor_izq = factor_izq[1:]
                    
        else:
                    
            signo_factor_izq = 1
            
        # multiplicamos
        
        for trozo in list_factores:
                
            if trozo[0] == '+':
                    
                signo_trozo = 1
                    
                trozo = trozo[1:]
                
            elif trozo[0] == '-':
                    
                signo_trozo = -1
                    
                trozo = trozo[1:]
                    
            else:
                    
                signo_trozo = 1
                
            var = signo_trozo*signo_factor_izq
                
            if var > 0:
                    
                signo = '+'
                
            else:
                    
                signo = '-'

            new_string += signo + factor_izq + trozo
        
        try:
            if element[j+1] == '*':

                put_in = '+('
                put_out = ')'
                
            else:
            
                put_in = ''
                put_out = ''
        
        except:
            
            put_in = ''
            put_out = ''
        
        if pos_inicial != 0:
        
            if element[:pos_inicial+1][-1] == '+' or element[:pos_inicial+1][-1] == '-':
                
                #print('opcion 1')
                
                element = element[:pos_inicial] + put_in + new_string + put_out + element[j+1:]
                
            else:
                
                #print('opcion 2')
                
                element = element[:pos_inicial+1] + put_in + new_string + put_out + element[j+1:]  

        else:
            
            #print('opcion 3')
            
            if pos_inicial-1 >= 0:
                
                init_string = element[:pos_inicial]
                
            else:
                
                init_string = ''
            
            element = init_string + put_in + new_string + put_out  + element[j+1:]
                
        #AQUI HAY Q CAMBIARLO

    elif evaluar(j+1,element):

        pos_final, factor_der = lado_der(i,j,element, arr_paren)
            
        if factor_der[0] == '+':
                    
            signo_factor_der = 1
                
            factor_der = factor_der[1:]
                
        elif factor_der[0] == '-':
                    
            signo_factor_der = -1
                
            factor_der = factor_der[1:]
                    
        else:
                    
            signo_factor_der = 1

        list_factores = factores(i,j,element)
        
        #print('FACTORES',list_factores)
        # multiplicamos

        primer_termino = True
            
        for trozo in list_factores:
                
            if trozo[0] == '+':
                    
                signo_trozo = 1
                    
                trozo = trozo[1:]
                
            elif trozo[0] == '-':
                    
                signo_trozo = -1
                    
                trozo = trozo[1:]
                    
            else:
                    
                signo_trozo = 1
                
            var = signo_trozo*signo_factor_der
                
            if var > 0:
                    
                signo = '+'
                
            else:
                    
                signo = '-'

                
            new_string += signo + trozo + factor_der
            
        element = element[:i] + '(' + new_string + ')'  + element[pos_final+1:]
        
        #HAY Q CORREGIR EL POS_FINAL, EL POS_FINAL CAMBIA SI ES UN SIGNO A SI ES UN PARENTESIS
    
    #print(element)
    element = element.replace('++','+')
    
    element = element.replace('+-','-')
    
    element = element.replace('-+','-')
    
    element = element.replace('--','+')
    
    return element

def find_interno(arr_paren,element):     #ENCUENTRA LOS PARENTESIS INTERNOS: de tipo ( jaskljaslkdjal) que no tienen () dentro
      
    paren_interno = False    
    
    for i in range(len(arr_paren)):
        
        if np.isnan(arr_paren[i]) == False:
            
            for j in range(i+1,len(arr_paren)):
                
                if arr_paren[j] == arr_paren[i]:
                    
                    break
                    
            paren_interno = True
                    
            for k in range(i+1,j-1):
                
                if np.isnan(arr_paren[k]) == False:
                    
                    paren_interno = False
                    
                    break
            
            if paren_interno == True:
                
                break
       
    if paren_interno == True:
        
        #print(i,j)
    
        element_new = multiplicar(i,j,element,arr_paren)

        if element_new != element:
            
            element = element_new
                
            arr_paren = parentesis_distribute(element)[0]
                
            element_new = find_interno(arr_paren,element)
            
            #print('Dentro del IF',element)
    else:
        
        element_new = element
    
    return element_new

def distribute(element):
    
    '''blablabla'''

    
    if '=' in element:

        raise(TensorSyntaxError("The element cannot contain an '=' sign."))

    element = delete_blanks(element)
    
    #print('delete_blanks',element)

    element = correct_multiplication(element)
    
    #print('correct multiplication',element)
    #print('INPUT:')
    #display(Math(element))
    
    arr_paren,cant_par = parentesis_distribute(element)
    
    arr_paren,element = check_parentesis(arr_paren,element)
    
    #print('parentesis', element)
    
    #print(arr_paren)
    
    element = find_interno(arr_paren,element)
    
    #print('find interno',element)
    
    element = element.replace('{','(')
    element = element.replace('}',')')    
    
    #print('FINAL', element)

    #print('OUTPUT:')
    #display(Math(element))
    
    return element


# Ejemplo:
#distribute('D(Tx[^j,_u]+D(Ty[^j],_u)*(C(Ty[^k],_k)+Tz[^b,_b]),_r)')




from .gr import (def_coords as coords, create_metric as metric, christoffel, riemann, ricci,
                 ricci_scalar as riccis, einstein, geodesic)

from .tensor import (setorder as order, series, new_var as var, new_con as con, new_fun as fun, new_ten as ten,
                    LeviCivita as lcivita, KroneckerDelta as kdelta, D, C ,display, set_space_time as spacetime,
                     simplify, expand, factor, det)



from .misc import setorder, series, new_var, new_con, new_fun, new_ten, set_space_time, factor_pt as factor,simplify_pt as simplify,expand_pt as expand
from .tensor import D, C, determinant as det
from .core import display
from .lcivita import LeviCivita
from .kdelta import KroneckerDelta


import re
import numpy as np
from collections import Counter
from copy import deepcopy
from .error import TensorSyntaxError
from .greek import greek_dict
from .distribute import distribute


def index_counter(element):
    
    #'Tg[_i,^j]*TC2[^i,_k,_l]+TC2[^i,_k,_l]-TC2[^i,_k,_l]*Tg[^l,_p]'
    
    # WE SPLIT THE STRING AND ASSIGN EVERY TERM TO A LIST
       
    list_strings = re.split('[- +]',element)
    
    if list_strings[0] == '':
        
        list_strings = list_strings[1:]
    
    # NOW WE SEARCH FOR EVERY STRING IN list_strings THE UPPER AND LOWER INDEX AND VERIFY IF THIS HAS SENSE

    error_indice_up_dn = False
    
    n_up = 0
    n_dn = 0
    
    contador_up = []
    contador_dn = []
    
    repeated_index = []
    
    n_termino = 0
    
    list_tensor_term_index = []
    
    indexacion_ok = True
    
    for string in list_strings:
        
        list_index = []
        
        n_up = 0
        n_dn = 0
    
        #'Tg[_i,^j]*TC2[^i,_k,_l]'

        # D(H[^l,_j],_k)
        for i in range(len(string)):
            
            if string[i] == '_' or string[i] == '^':
                
                j = i

                while string[j] != ',' and string[j] != ']' and string[j] != ')':

                    j += 1

                #list_index.append(string[i:i+2])
                list_index.append(string[i:j])

        
        list_tensor_term_index.append(list_index)
        
        #['_i', '^j', '^i', '_k', '_l']
        
        # WE MUST VERIFY IF EVERY ELEMENT OF THIS ARRAY HAS ONLY TWO CHARS BY STRING. THEN
            
        up_dn_index_list = [] 

        symbol_index_list = []

        for letras in list_index:
        
            up_dn_index_list.append(letras[0])
            #['_', '^', '^', '_', '_']
            
            symbol_index_list.append(letras[1:])
            #['i', 'j', 'i', 'k', 'l']

        for i in range(len(symbol_index_list)):
            
            contador = 1
            
            for j in  [x for x in range(len(symbol_index_list)) if x != i]:
                
                if symbol_index_list[i] == symbol_index_list[j]:
                    
                    contador += 1
                    
                    if contador == 2:
            
                        if not ((up_dn_index_list[i] == '_' and up_dn_index_list[j] == '^') or (up_dn_index_list[i] == '^' and up_dn_index_list[j] == '_')):
                            
                            #print('not up or dn')
                            
                            indexacion_ok = False
                    
                    elif contador > 2:
                        
                        #print('repetido mas de 2 veces')
                        
                        #print(string,up_dn_index_list,symbol_index_list)
                        
                        indexacion_ok = False
          
        counter=Counter(symbol_index_list)
        
        arreglo_momentaneo = [counter[x] for x in counter.keys()]
        
        #AHORA HAY QUE CONTAR CUANTOS INDICES HAY SIN REPETIR
        
        arreglo_dict = np.asarray([(k,v) for k,v in counter.items()]) 
        #THIS LINE CONVERTS THE DICTIONARY TO A LIST, AND THIS TO AN ARRAY!!!!!!!
        
        for i in range(len(counter.items())):
        
            if arreglo_dict[i,1] == '1':  
                # SI EL ELEMENTO i-ESIMO NO SE REPITE (SE CUENTA UNA SOLA VEZ)
               
                pos = np.where(np.asarray(symbol_index_list) == arreglo_dict[i,0] )[0][0]
                         
                el_simbolo = up_dn_index_list[pos]    
                #EVALUA EL ARREGLO DE ^_ EN LA POSICION
                # DONDE EL ELEMENTO QUE SOLO APARECE UNA SOLA VEZ TIENE 
                #EL MISMO SIMBOLO EN EL DICCIONARIO QUE EN EL symbol_index_list
                
                if el_simbolo == '^':
                    
                    n_up += 1
                             
                elif el_simbolo == '_':
                    
                    n_dn += 1
                               
                else:
                    
                    error_indice_up_dn = True
            
            else:
                
                # SI EL INDICE ESTA REPETIDO EL CONTADOR DEL DICCIONARIO SERA 2, SI SE REPITE MAS VECES SE RETORNARA
                # EL ERROR EN INDICES REPETIDOS ANTES COMO: indexacion_ok
                # GUARDAMOS LOS INDICES QUE ESTAN REPETIDOS, SOLO EL SIMBOLO, NO EL INDICADOR _ O ^ Y LO GUARDAMOS PARA
                # CADA TERMINO DE LA ECUACION (CADA TERMINO: PARA CADA ELEMENTO SEPARADO POR * )
                
                repeated_index.append([n_termino,arreglo_dict[i,0]])  #GUARDAMOS LOS INDICES REPETIDOS
        
        if all(value == 1 for value in  counter.values()) == True:
            
               
            repeated_index.append([n_termino,'none'])  #GUARDAMOS LOS INDICES REPETIDOS
      
        contador_up.append(n_up)
        contador_dn.append(n_dn)
        
        
        n_termino +=1    
    
        #DE ESTOS INDICES, CUANTOS SON UP Y CUANTOS DOWN, Y Q ESTO SEA CONSISTENTE PARA TODOS LOS TERMINOS
        #RETORNAR SI ES CIERTO O NO
    
    # AHORA RETORNA EL element INICIAL (FORMULA COMPLETA) CORTADA EN STRINGS DONDE 
    # CADA UNO SE SUMA SOBRE SUS PROPIOS INDICES REPETIDOS
    
    indice_up_dn_OK = not error_indice_up_dn

    
    return (indexacion_ok,indice_up_dn_OK,contador_up,contador_dn,repeated_index,list_tensor_term_index)
        

def writelatex(element,sum_over_diz_index,condition):


    element = checknames(element) # Cambia los nombres para mostrarlos correctamentes si son tensores tipicos (Ricci, Christoffel, etc.)
    
    pattern = '([\+ \-]+)'
    list_strings = [i.strip() for i in re.split(pattern, element) if i.strip()]
    #GUARDA LOS SIGNOS Y LOS TENSORES, TODO EN UN SOLO ARREGLO
    

    
    #'Tg[_i,^j,_u] TC2[^i,_k,_l] Tw[^u] +TC2[^j,_k,_l] - TC2[^j,_q , _w ]*Tg[^q,_l,_k] Tw[^w]+ D(Te[^j,_k],_r)*Tg[^r,_l,_u]* Tw[^u]'
     
    n_termino = 0
    
    new_element = []

    for string in list_strings:

        if string !="+" and string !="-":        
            
            string = string_to_latex(string)
      
            #print('string desp de string to latex',string)

            #'LO QUE VUELVE DESPUES DE TRANSFORMARLO A LATEX'
            #AQUI VUELVE EN LATEX EL TERMINO ENTERO 'TC2[^j,_q , _w ]*Tg[^q,_l,_k] Tw[^w]' SIN LOS SIGNOS + O -
        
            if condition == 'Y':

                new_element.append('\sum_{')

                for i in range(len(sum_over_diz_index)):

                    if sum_over_diz_index[i][0] == n_termino:

                        new_element.append(sum_over_diz_index[i][1])
                        new_element.append(',')

                new_element.pop()
                new_element.append('}')
                
            n_termino += 1 

        new_element.append(string)
  
    new_math_element = ''.join(new_element)
  


    new_math_element = new_math_element.replace("\sum_{none}","")  
    #elimina las sumas sobre indices inexistentes
    
    new_math_element = new_math_element.replace("*","\cdot ")
    #reemplaza los * por signos bonitos de multiplicacion

    #print('new math element',new_math_element)    

    return new_math_element   



def non_repeated_i(element):
    
    examine(element)

    element = distribute(element)
    
    r1,r2,r3,r4,r5,r6 = index_counter(element)

    non_repeated_indexes = []

    for term in r6:
    
        for index in term:
            
            if index[1] != r5[0][1]:
            
                non_repeated_indexes.append(index)

    return non_repeated_indexes

def add_examine(left,right):

    if set(left) != set(right):

        # Deberia mostrar los indices que estan incorrectos

        raise(TensorSyntaxError('Bad index Definition'))

def mul_examine(rep):

    for i,index in enumerate(rep):

        name = index[1:] # 'i'

        for j in range(i+1,len(rep)):

            #print('rep[j]',rep[j])

            if name == rep[j][1:]:

                if rep[j][0] == index[0]:

                    #print('name,rep[j][1:]',name,rep[j][1:])

                    #print('index[0],rep[j][0]',index[0],rep[j][0])

                    raise(TensorSyntaxError('Bad'))
    
def der_examine(b):

    if len(b)<2:

        raise(TensorSyntaxError('Syntax Error.'))

    elif b[0] != '_' and b[0] != '^':

        raise(TensorSyntaxError('Wrong symbol.'))

    elif b.count(',') !=  0:

        raise(TensorSyntaxError('It must be only one index.'))

    elif b.count('_') + b.count('^') > 1:

        raise(TensorSyntaxError("Syntax Error. Problem with the '_' or '^' symbols."))

def syntax(string,rank):

    '''
    This function verifies if the syntax for the requested tensor is correct.
    '''
    
    lista = string.split(',')

    for i in lista:
        
        if len(i) < 2:
            
            raise(TensorSyntaxError("Bad index Definition: Missing index or symbol."))
        
        elif i[0] != '_' and i[0] != '^':
            
            raise(TensorSyntaxError("Bad index Definition: Missing a '_' or a '^'."))
            
        elif i.count('_') + i.count('^') > 1:
            
            raise(TensorSyntaxError("Bad index Definition: Missing a ',' or typo error."))

    if rank < len(lista):

        raise(TensorSyntaxError("Too many indices for this tensor."))

    elif rank > len(lista):

        raise(TensorSyntaxError("Too few indices for this tensor."))




import numpy as np 
import tqdm
from itertools import product as iterprod 

from pytearcat.tensor.misc import new_ten,reload_all
from pytearcat.tensor.core import config
from pytearcat.tensor.tensor import tensor_series, Tensor

from .christoffel import calculate_christoffel, D
from .riemann import calculate_riemann

from pytearcat.tensor.core.core import core_calc,display_IP,Latex_IP

if core_calc == 'sp':

    from pytearcat.tensor.core.core import diff, sympify, factor

elif core_calc == 'gp':

    from pytearcat.tensor.core.core import diff, simplify

def calculate_ricci(All = False):

    # Ricci Tensor

    global Ricci


    if config.ricci is None:

        Ricci = config.create_ten('Ricci',Tensor('Ricci',2))

        config.ricci = Ricci

    else: 

        Ricci = config.ricci


    if config.riemann is None:

        Riemann_local = calculate_riemann()

    else:

        Riemann_local = config.riemann
   


    if Ricci.indices[0] == False:

        Ricci_list = np.full((config.dim,config.dim), False, dtype=bool)

        display_IP(Latex_IP(r'Ricci Tensor $R_{\alpha \beta}$'))

        for p in tqdm.tqdm_notebook(iterprod(range(config.dim),repeat=2),total=config.dim**2):

            i = p[0]
            j = p[1]


            if Ricci_list[i][j] == False:

                Temp_Ricci = 0

                for k in range(config.dim):

                    Temp_Ricci += Riemann_local.tensor[8][k][i][k][j]

                if config.ord_status == True:

                    Ricci.tensor[0][i][j] = tensor_series(Temp_Ricci)

                else:

                    if core_calc == 'sp':

                        Ricci.tensor[0][i][j] = sympify(factor(Temp_Ricci))

                    elif core_calc == 'gp':

                        Ricci.tensor[0][i][j] = simplify(Temp_Ricci)

                # Simetrias

                Ricci_list[i][j] = True

                Ricci.tensor[0][j][i] = Ricci.tensor[0][i][j] # Simetria de Ricci

                Ricci_list[j][i] = True

        Ricci.indices[0] = True

    else:

        display_IP(Latex_IP(r"Ricci Tensor $R_{\alpha \beta}$ already calculated"))

    if All == True:

        Ricci.complete('_,_')

    Ricci.space()

    return Ricci



def calculate_ricci_scalar():

    global Ricci_Scalar

    if config.ricci is None:

        Ricci_local = calculate_ricci()
        
    else:

        Ricci_local = config.ricci

    if config.ricciS is None:

        Ricci_Scalar = 0

        for p in tqdm.tqdm_notebook(iterprod(range(config.dim),repeat=2),total=config.dim**2):

            i = p[0]
            j = p[1]

            Ricci_Scalar += config.g.tensor[3][i][j] * Ricci_local.tensor[0][i][j]

        config.ricciS = Ricci_Scalar
    
    else:

        display_IP(Latex_IP(r"Ricci Scalar $R$ already calculated"))

        return config.ricciS

    display_IP(Latex_IP(r'Ricci Scalar $R$'))

    if config.ord_status == True:

        Ricci_Scalar = tensor_series(Ricci_Scalar)


    else:

        if core_calc == 'sp':

            Ricci_Scalar = sympify(factor(Ricci_Scalar))

        elif core_calc == 'gp':

            Ricci_Scalar = simplify(Ricci_Scalar)

    return Ricci_Scalar


from IPython.display import display as display_IP, Math as Math_IP, Latex as Latex_IP
#from platform import system
import pkg_resources

__required = {'jupyter','numpy' ,'sympy', 'tqdm','giacpy'}
__installed = {pkg.key for pkg in pkg_resources.working_set}
__missing = __required - __installed

if len(__missing) != 0:

    if 'giacpy' in __missing and len(__missing) == 1:

        pass

    else:

        raise(EnvironmentError("There are missing modules:"(__missing-{'giacpy'})))


if 'giacpy' in __installed:

    core_calc = 'gp'

else:

    core_calc = 'sp'

if core_calc == 'gp':

    from .gpwrap import *
    import io
    from contextlib import redirect_stdout
    from sympy import simplify as sp_simplify, sympify as sp_sympify, latex as sp_latex, Array as sp_Array

    def get_name(element):
        
        f = io.StringIO()
        with redirect_stdout(f):
            print(element)
        element = f.getvalue()

        return element[:-1]


    def tolatex(element):

        f = io.StringIO()
        with redirect_stdout(f):
            print(latex(element))
        element = f.getvalue()

        string =  element[1:-2]

        return string

elif core_calc == 'sp':

    from sympy import *
    
    def get_name(f):

        return f.name



import numpy as np 
from re import findall,search
from .core import config
from .core import ptseries
from .core.core import get_name, simplify, factor, expand, series as Series,core_calc
from .core import core
from .core.error import TensorSyntaxError
from .tensor import Tensor

if core_calc == 'gp':

    import io
    from contextlib import redirect_stdout

def factor_pt(x):

    if core_calc == 'gp':

        iscoreobj = isinstance(x,core.gpcore)

    elif core_calc == 'sp':

        iscoreobj =  isinstance(x,core.Expr)

    isnumber = False

    if isinstance(x,int) or isinstance(x,float) or isinstance(x,np.number):

        isnumber = True

    elif isinstance(x,Tensor):

        return x.factor()

    elif iscoreobj or isnumber:

        return factor(x)

    else:

        raise TypeError("The arg is not a mathematical object.")


def simplify_pt(x):

    if core_calc == 'gp':

        iscoreobj = isinstance(x,core.gpcore)

    elif core_calc == 'sp':

        iscoreobj =  isinstance(x,core.Expr)

    isnumber = False

    if isinstance(x,int) or isinstance(x,float) or isinstance(x,np.number):

        isnumber = True

    elif isinstance(x,Tensor):

        return x.simplify()

    elif iscoreobj or isnumber:

        return simplify(x)

    else:

        raise TypeError("The arg is not a mathematical object.")

def expand_pt(x):

    if core_calc == 'gp':

        iscoreobj = isinstance(x,core.gpcore)

    elif core_calc == 'sp':

        iscoreobj =  isinstance(x,core.Expr)

    isnumber = False

    if isinstance(x,int) or isinstance(x,float) or isinstance(x,np.number):

        isnumber = True

    elif isinstance(x,Tensor):

        return x.expand()

    elif iscoreobj or isnumber:

        return expand(x)

    else:

        raise TypeError("The arg is not a mathematical object.")

    

def set_space_time(x=True):

    config.space_time = x

def not_intersection(L1,L2):
    '''
    example: not_intersection(conf_vars,new_vars)

    retorna las variables de L2 que no estan en L1
    '''
    return [i for i in L2 if i not in L1]

def intersection(L1,L2):
    '''
    example: not_intersection(conf_vars,new_vars)

    retorna las variables de L2 que estan en L1
    '''
    return [i for i in L2 if i in L1]

def setorder(var,n):

    '''
    Sets the expansion order equal to "n" around variable "var". 

    Note: 
    This sets the expansion order equal to "n", allowing the program to truncate
    all the calculations up to this order. You can also revert to a value lower
    than "n" without having to restart the kernel. However, this does not work
    if the new order is greater than the first defined order since all the 
    higher-order terms were discarded in the first execution of this function.

    '''

    if core_calc == 'gp' and not isinstance(var,core.gpcore):

        raise(TypeError("var must be a giacpy object."))

    elif core_calc == 'sp' and not isinstance(var,core.Expr):

        raise(TypeError("var must be a sympy object."))

    ptseries.ord_status = True
    config.ord_status = True

    ptseries.ord_var = var
    config.ord_var = var

    ptseries.ord_n = n
    config.ord_n = n

def series(element):

    '''
    Compute the series of an element.
    '''

    if core_calc == 'gp':

        string = "simplify(Series(element, x = config.ord_var, n = config.ord_n+1))"

    elif core_calc == 'sp':

        string = "simplify(expand(Series(element, x = config.ord_var, n = config.ord_n+1)))"

    result = eval(string,locals(),globals())

    return result

def new_var(*args):

    '''
    example: new_var('x','y',...)
    
    '''

    input_vars = list(args)

    if not all(isinstance(var, str) for var in input_vars):

        raise(TypeError("Arguments must be str."))

    if 'epsilon' in input_vars or 'Epsilon' in input_vars:

        raise(NameError("Pytearcat does not support 'epsilon' as a variable name. Please use another name."))

    if 'varepsilon' in input_vars or 'Varepsilon' in input_vars:

        raise(NameError("Pytearcat does not support 'varepsilon' as a variable name. Please use another name."))

    names = []

    for i in config.var:
        names.append(get_name(i))

    
    new_vars = not_intersection(names,input_vars)
    old_vars = intersection(names,input_vars)

    variables = []
    def_variables = [i for i in config.var if get_name(i) in old_vars]

    if len(new_vars) != 0:

        for i in new_vars:

            variables.append(config.create_var(i))
    
    if len(variables) == 1:

        return variables[0]

    elif len(variables) == 0:

        if len(def_variables) == 1:

            return def_variables[0]
        else:
            return def_variables

    else:
        
        return variables

def new_con(*args):

    input_const = list(args)

    if not all(isinstance(var, str) for var in input_const):

        raise(TypeError("Arguments must be str."))

    if 'epsilon' in input_const or "Epsilon" in input_const:

        raise(NameError("Pytearcat does not support 'epsilon' as a variable name. Please use another name."))

    if 'varepsilon' in input_const or "Varepsilon" in input_const:

        raise(NameError("Pytearcat does not support 'varepsilon' as a variable name. Please use another name."))

    names = []

    for i in config.con:
        names.append(get_name(i))
    
    new_const = not_intersection(names,input_const)
    old_const = intersection(names,input_const)

    constants = []
    def_constants = [i for i in config.con if get_name(i) in old_const]

    if len(new_const) != 0:

        for i in new_const:

            constants.append(config.create_con(i))
    
    if len(constants) == 1:

        return constants[0]

    elif len(constants) == 0:

        print('Constant(s) already defined.')

        if len(def_constants) == 1:

            return def_constants[0]
        else:
            return def_constants

    else:
        
        return constants

def new_fun(fun_symbol,var_symbol,overwrite=False):

    '''
    Defines a new function receiving the name and the variables respectively. new_fun('f','t,x,y')
    '''

    # Aca se puede definir con un kwargs para que entren n funciones de la forma (f,variables)

    if not isinstance(fun_symbol,str):

        raise(TypeError("Function symbol must be a string"))

    if 'epsilon' in fun_symbol or "Epsilon" in fun_symbol:

        raise(NameError("Pytearcat does not support 'epsilon' as a variable name. Please use another name."))

    if 'varepsilon' in fun_symbol or "Varepsilon" in fun_symbol:

        raise(NameError("Pytearcat does not support 'varepsilon' as a variable name. Please use another name."))

    existing_names = []

    for i in config.fun:

        existing_names.append(get_name(i))

    new_var(*var_symbol.split(',')) # creates the variables if they not exist

    if fun_symbol not in existing_names or (fun_symbol in existing_names and overwrite == True) :

        function = config.create_fun(fun_symbol,var_symbol)

    else:

        print('Function %s(%s) is already defined'%(fun_symbol,var_symbol))

        fun_index = existing_names.index(fun_symbol)

        function = config.fun[fun_index]

    return function

def new_ten(tname,index):

    if len(findall('([^A-Za-z0-9])',tname)) != 0:

        raise TensorSyntaxError('The tensor name cannot contain any special character.')

    if tname in config.default_tensors:

        raise NameError("%s is protected as a default tensor name."%tname)

    if 'epsilon' == tname or "Epsilon" == tname:

        raise(NameError("Pytearcat does not support 'epsilon' as a variable name. Please use another name."))

    if 'varepsilon'  == tname or "Varepsilon"  == tname:

        raise(NameError("Pytearcat does not support 'varepsilon' as a variable name. Please use another name."))

    A = Tensor(tname,index)


    exec(reload_all('config'))

    return config.create_ten(tname,A)

def new_temp(temp_name,t_obj):

    return config.create_temp(temp_name,t_obj)

def delete_temp():

    for i in config.temp_name:

        exec('del(config.%s)'%i,locals(),globals())

        config.__all__.remove(i)

    
    config.temp = []

    config.temp_name = []


def in_var(name):

    var_list = list(map(get_name,config.var))

    if name in var_list:

        return True

    else:

        return False

def in_fun(name):

    fun_list = list(map(get_name,config.fun))

    if name in fun_list:

        return True

    else:

        return False

def in_ten(name):

    ten_list = list(map(get_name,config.ten))

    if name in ten_list:

        return True

    else:

        return False

def in_temp(name):

    temp_list = config.temp_name

    if name in temp_list:

        return True

    else:

        return False


def name_handler(string, name):

    # Revisar esto, deberia ser con un for i in config.var revisar si config.var[i].name == name.

    

    var_list = list(map(get_name,config.var))

    fun_list =list(map(get_name,config.fun))

def reload_all(new_module):

    ''' 

    This function reloads all the variables in __all__ from module
        
    Thanks to hurturk on https://stackoverflow.com/questions/44492803/python-dynamic-import-how-to-import-from-module-name-from-variable?newreg=9312458d429647fc8ec2a72e5655d197
        
    '''

    string = "globals().update({{n: getattr(module, n) for n in module.__all__}} if hasattr(module, '_all_') else {k: v for (k, v) in module.__dict__.items() if not k.startswith('_')})"

    string = string.replace('module',new_module)

    return string


from .core import *
from pytearcat.tensor.core import config
from itertools import product as iterprod
from .interp import add_examine, mul_examine

def reload_all(new_module):

    ''' x

    This function reloads all the variables in __all__ from module
        
    Thanks to hurturk on https://stackoverflow.com/questions/44492803/python-dynamic-import-how-to-import-from-module-name-from-variable?newreg=9312458d429647fc8ec2a72e5655d197
        
    '''

    string = "globals().update({{n: getattr(module, n) for n in module.__all__}} if hasattr(module, '_all_') else {k: v for (k, v) in module.__dict__.items() if not k.startswith('_')})"

    string = string.replace('module',new_module)

    return string

def out_intersection(L1,I1,L2,I2):
    '''
    example: not_intersection(conf_vars,new_vars)

    retorna las variables que no son compartidas en L1 y L2
    '''
    L = L1+L2
    
    I = I1+I2

    lista_intersection = [i for i in L2 if i in L1]

    lista = [x for x in L if x not in lista_intersection]

    out = {lista[i] : i for i in range(len(lista))}

    ind_dict =  dict((k,i) for i,k in enumerate(L))

    ind_out = [ I[ind_dict[x]] + L[ind_dict[x]] for x in out ]

    string = ','.join(ind_out)

    return out, string

def in_intersection(L1,I1,L2,I2):

    '''
    example: in_intersection(conf_vars,new_vars)  ['^i','_j'] ['_i','^l']

    ['i','j','i','l']

    retorna un diccionario con la interseccion
    '''

    L = L1+L2
    
    I = I1+I2

    lista_intersection = [i for i in L2 if i in L1]

    in_dict = {lista_intersection[i] : i for i in range(len(lista_intersection))}

    full_rep = []

    for i in range(len(L)):

        if L[i] in lista_intersection:

            full_rep.append('%s%s'%(I[i],L[i]))

    return in_dict, full_rep

def construct(elem,dim,n):
    
    '''
    It builds a nested list filled with elem on every item. 
    It returns an object with shape (dim,dim,dim,...,dim) of n entries.
    elem must be a string and you can recover each elem by using result[i][j][k] with i,j,k integers < dim
    '''
    
    string = ('%s,'%elem)*dim

    string = '['+string[:-1]+'],'

    for k in range(n-1):

        string = '['+(string*dim)[:-1]+'],' 
    
    return eval(string[:-1])

class Tdata:

    '''
    A class to represent the data of a Tensor.
    

    ...

    Attributes 
    -----------

    str_index : str
        String indicating the corresponding index combination of the Tensor, i.e. "^i,_j"

    elements : list or array
        list containing the data elements of the corresponding index combination.

    Methods
    ----------
    factor():
        Factors the Tensor data
    
    simplify():
        Simplifies the Tensor data
    

    '''

    def __init__(self,str_index,elements):

        
        self.full_index = str_index # '^i,_j' 

        self.full_list = self.full_index.split(',')
        
        self.updn = [symbol[0] for symbol in self.full_list]
            
        self.index_names = [symbol[1:] for symbol in self.full_list]
            
        self.elements = elements # T

    def __getitem__(self,item):

        return self.elements[item]

    def auto_sum(self):

        '''
        recibe tdata

        se fija en el index_names y busca los indices repetidos. Suma sobre si mismo una sola vez.
        Un solo indice repetido

        '''

        exec(reload_all('config'),locals(),globals())

        if config.space_time == True:

            dim = config.dim

        else:

            dim = config.dim - 1

        # AGREGAR UN EXAMINE PARA AUTO_SUM

        lista = self.index_names

        rep = False

        #print('en el auto_sum',self.full_list)

        for i,name_i in enumerate(lista):

            for j,name_j in enumerate(lista):

                if name_i == name_j and i != j:

                    pos_i,pos_j = i,j

                    rep = True
                    #print(pos_i,pos_j,name_i,name_j)

                    break

        self_index = ''

        return_index = ''

        full_index = ''

        k = 0

        if rep == True: # Si encuentra un indice repetido

            for i in range(len(self.updn)):

                if i == pos_i or i == pos_j:
                
                    self_index += '[q]'
                
                else:
                    
                    self_index += '[p[%d]]'%k

                    return_index += '[p[%d]]'%k

                    full_index += '%s,'%self.full_list[i]

                    k += 1

            full_index = full_index[:-1]

            return_tensor = construct(0,dim,len(self.updn)-2)

            for p in iterprod(range(dim),repeat=len(self.updn)-2): # for para ir asignarlo

                temp = 0

                for q in range(dim):

                    string = 'self.elements%s'%(self_index)

                    temp += eval(string,locals(),globals())
                
                string = 'return_tensor%s = temp'%return_index

                exec(string,locals(),globals())
                
            return Tdata(full_index,return_tensor)   #se retorna como el self

        else:

            return self

    def __neg__(self):

        return (-1) * self 

    def __pos__(self):

        return self

        
    def __mul__(self,other):

        exec(reload_all('config'),locals(),globals())

        if config.space_time == True:

            dim = config.dim

        else:

            dim = config.dim - 1
        
        if isinstance(other,Tdata):
        
            rep_index, rep_list = in_intersection(self.index_names,self.updn,other.index_names,other.updn)

            mul_examine(rep_list)

            new_index, return_string = out_intersection(self.index_names,self.updn,other.index_names,other.updn)

            new_rank = len(new_index)

            if new_rank != 0:
                
                return_tensor = construct(0,dim,new_rank)

            if len(rep_index) != 0: # Si hay indices repetidos

                new_iterstr = ''

                for i in range(new_rank):

                    new_iterstr += '[p[%d]]'%i

                self_iterstr = ''

                for i, name in enumerate(self.index_names):

                    if name in rep_index:

                        self_iterstr += '[q[%d]]'%rep_index[name]

                    else:

                        self_iterstr += '[p[%d]]'%new_index[name]

                other_iterstr = ''

                for i, name in enumerate(other.index_names):

                    if name in rep_index:

                        other_iterstr += '[q[%d]]'%rep_index[name]

                    else:

                        other_iterstr += '[p[%d]]'%new_index[name]

                if new_rank != 0:

                    for p in iterprod(range(dim),repeat=new_rank): # for para ir asignarlo

                        temp = 0

                        for q in iterprod(range(dim), repeat = len(rep_index)):

                            string = 'self.elements%s*other.elements%s'%(self_iterstr,other_iterstr)

                            temp += eval(string,locals(),globals())

                        string = 'return_tensor%s = temp'%new_iterstr

                        exec(string,locals(),globals())
                
                else:

                    temp = 0

                    for q in iterprod(range(dim), repeat = len(rep_index)):

                        string = 'self.elements%s*other.elements%s'%(self_iterstr,other_iterstr)

                        temp += eval(string,locals(),globals())

                    return temp

            else:

                return_string = '%s,%s'%(self.full_index,other.full_index)

                new_iterstr = ''

                for i in range(new_rank):

                    new_iterstr += '[p[%d]]'%i

                self_iterstr = ''

                k = 0

                while k < len(self.updn):

                    self_iterstr += '[p[%d]]'%k

                    k += 1

                other_iterstr = ''

                while k < len(self.updn)+len(other.updn):

                    other_iterstr += '[p[%d]]'%k
                    
                    k += 1

                for p in iterprod(range(dim),repeat=new_rank): # for para ir asignarlo

                    temp = 0
                    
                    string = 'self.elements%s*other.elements%s'%(self_iterstr,other_iterstr)

                    temp = eval(string,locals(),globals())

                    string = 'return_tensor%s = temp'%new_iterstr

                    exec(string,locals(),globals())

            return  Tdata(return_string,return_tensor)            

        else: # Not Tdata * Tdata

            same_rank = len(self.updn)
            
            return_tensor = construct(0,dim,same_rank)

            index = ''

            for i in range(same_rank):

                index += '[p[%d]]'%i

            for p in iterprod(range(dim),repeat = same_rank):

                string = 'return_tensor%s = self.elements%s*other'%(index,index)

                exec(string,locals(),globals())

            return Tdata(self.full_index,return_tensor)
        
    def __rmul__(self,other):

        exec(reload_all('config'),locals(),globals())

        if config.space_time == True:

            dim = config.dim

        else:

            dim = config.dim - 1

        same_rank = len(self.updn)
            
        return_tensor = construct(0,dim,same_rank)

        index = ''

        for i in range(same_rank):

            index += '[p[%d]]'%i

        for p in iterprod(range(dim),repeat = same_rank):

            string = 'return_tensor%s = self.elements%s*other'%(index,index)

            exec(string,locals(),globals())

        return Tdata(self.full_index,return_tensor)

    def __add__(self,other):

        exec(reload_all('config'),locals(),globals())

        if config.space_time == True:

            dim = config.dim

        else:

            dim = config.dim - 1

        add_examine(self.full_index,other.full_index)

        self_lista = self.full_index.split(',')
        other_lista = other.full_index.split(',')

        if set(self_lista) == set(other_lista): 

            self_index = ''
            other_index = ''

            for i in range(len(self.updn)):

                self_index += '[p[%d]]'%i

            for j in range(len(other.updn)):

                i = 0

                while other_lista[j][1:] != self_lista[i][1:]:

                    i += 1

                other_index += '[p[%d]]'%i

            return_tensor = construct(0,dim,len(self.updn))

            for p in iterprod(range(dim),repeat=len(self.updn)): # for para ir asignarlo

                temp = 0

                string = 'self.elements%s+other.elements%s'%(self_index,other_index)

                temp = eval(string,locals(),globals())

                string = 'return_tensor%s = temp'%self_index   # se ordena como el self

                exec(string,locals(),globals())
            
            return Tdata(self.full_index,return_tensor)   #se retorna como el self

        else:

            print('ERROR')

    def __sub__(self,other):

        exec(reload_all('config'),locals(),globals())

        if config.space_time == True:

            dim = config.dim

        else:

            dim = config.dim - 1

        self_lista = self.full_index.split(',')
        other_lista = other.full_index.split(',')

        if set(self_lista) == set(other_lista): 

            self_index = ''
            other_index = ''

            for i in range(len(self.updn)):

                self_index += '[p[%d]]'%i

            for j in range(len(other.updn)):

                i = 0

                while other_lista[j][1:] != self_lista[i][1:]:

                    i += 1

                other_index += '[p[%d]]'%i

            return_tensor = construct(0,dim,len(self.updn))

            for p in iterprod(range(dim),repeat=len(self.updn)): # for para ir asignarlo

                temp = 0

                string = 'self.elements%s - other.elements%s'%(self_index,other_index)

                temp = eval(string,locals(),globals())

                string = 'return_tensor%s = temp'%self_index   # se ordena como el self

                exec(string,locals(),globals())
            
            return Tdata(self.full_index,return_tensor)   #se retorna como el self

        else:

            print('error')

    def __truediv__(self,other):
        
        exec(reload_all('config'),locals(),globals())

        if config.space_time == True:

            dim = config.dim

        else:

            dim = config.dim - 1
        
        if not isinstance(other,Tdata):

            same_rank = len(self.updn)
            
            return_tensor = construct(0,dim,same_rank)

            index = ''

            for i in range(same_rank):

                index += '[p[%d]]'%i

            for p in iterprod(range(dim),repeat = same_rank):

                string = 'return_tensor%s = self.elements%s/other'%(index,index)

                exec(string,locals(),globals())

            return Tdata(self.full_index,return_tensor)

    def factor(self):

        '''
        Factor the Tensor data.
        '''

        self.elements = sympify(factor(self.elements))

    def simplify(self):

        '''
        Simplify the Tensor data.
        '''

        self.elements = sympify(simplify(np.array(self.elements)).tolist())



import numpy as np
from itertools import permutations, product as iterprod
from .tensor import ordenar, compare, Tensor
from .core import config
from .core.core import *
from .core.display import display
from .core.interp import syntax
from .core.tdata import Tdata, construct

if core_calc == 'gp':

    import io
    from contextlib import redirect_stdout
    from .core.display import gp_pretty_latex
    from .core.core import sp_simplify, sp_sympify, sp_latex, sp_Array



class KroneckerDelta():

    
    def __init__(self):

        dim = config.dim

        self.name = 'KroneckerDelta'

        self.n = 2

        self.sequence = ordenar(self.n)

        self.indices = np.full((2**self.n), False)

        if core_calc == 'sp':

            string = 'sympify(np.nan),'*dim

        elif core_calc == 'gp':
            
            string = 'np.nan,'*dim

        string = '['+string[:-1]+'],'

        for k in range(self.n-1):

            string = '['+(string*dim)[:-1]+'],'  

        string = 'self.tensor = [' + (string*(2**self.n))[:-1]+']'

        exec(string)# primer casillero corresponde al indice de la combinacion de indices (^,_,_,^,_), son 2**n combinaciones, los demas son primer indice, segundo indice, tercer indice, .... del tensor cada uno de ellos va desde 0 a dim, donde hay n dimnesiones

        #----

        if core_calc == 'sp':

            string = 'sympify(np.nan),'*(dim-1)

        elif core_calc == 'gp':
            
            string = 'np.nan,'*(dim-1)

        string = '['+string[:-1]+'],'

        for k in range(self.n):

            string = '['+(string*(dim-1))[:-1]+'],'  

        string = 'self.tensor_sp = [' + (string*(2**(self.n)))[:-1]+']'

        exec(string)# primer casillero corresponde al indice de la combinacion de indices (^,_,_,^,_), son 2**n combinaciones, los demas son primer indice, segundo indice, tercer indice, .... del tensor cada uno de ellos va desde 0 a dim, donde hay n dimnesiones

        # ----------------------------------------

        if core_calc == 'sp':

            Eye = eye(dim)

            self.__assign(Eye.tolist(),'_,_')
            self.__assign(Eye.tolist(),'^,_')
            self.__assign(Eye.tolist(),'_,^')
            self.__assign(Eye.tolist(),'^,^')
        
        elif core_calc == 'gp':

            Eye = idn(dim)

            self.__assign(list(Eye),'_,_')
            self.__assign(list(Eye),'^,_')
            self.__assign(list(Eye),'_,^')
            self.__assign(list(Eye),'^,^')

    def __repr__(self):

        return '%s Symbol defined'%self.name

    def __str__(self):

        if self.n <=2:

            self.display()

            string = ''

        else:

            string = 'Too many tensor indices. Please use the display method.'
        
        return string

    def __scalar_call(self,updn,numbers,Nindex,ten_call):
    
        tdat_str = ''
        
        new_string_right = ''
        
        for i in numbers:
                
            new_string_right += '[%s]'%i
        
        return eval('self.%s[%s]%s'%(ten_call,Nindex,new_string_right))


    def __Tdata_call(self,updn,coord,numbers,Nindex,dim,ten_call):
    
        tdat_str = ''
        
        new_string_right = ''
        
        k = 0
        j = 0
        
        for i in range(len(updn)):
            
            if coord[i] not in numbers:
            
                tdat_str += updn[i] + coord[i] + ','
                
                new_string_right += '[p[%d]]'%k
                
                k += 1
                
            else: 
                
                new_string_right += '[%s]'%numbers[j]
                
                j += 1

        tdat_str = tdat_str[:-1]
                
                    
        new_rank = len(coord)-len(numbers)
        
        elements = construct(0,dim,new_rank)
        
        new_string_left = ''
        
        for k in range(new_rank):

            new_string_left += '[p[%d]]'%k
        
        
        for p in iterprod(range(dim), repeat = new_rank):

            exec_str = 'elements%s = self.%s[%s]%s'%(new_string_left,ten_call,Nindex,new_string_right)
        
            exec(exec_str,locals(),globals())

        data_result = Tdata(tdat_str,elements)

        TEMP = Tensor('TEMP',new_rank)

        TEMP.assign(data_result,tdat_str,printing=False)

        return TEMP(tdat_str)

    def __call__(self,str_index):

        '''
        Solo acepta a lo mas la contraccion de 1 solo indice. Hay que incluir esto
        en el examine de este call

        Funcionando para space only
        '''

        syntax(str_index,self.n)

        if config.space_time == 1:

            dim = config.dim

            ten_call = 'tensor'

        else:

            dim = config.dim - 1

            ten_call = 'tensor_sp'

        lista = str_index.split(',')

        updn = [symbol[0] for symbol in lista]

        coord = [symbol[1:] for symbol in lista]
        
        Nindex = compare(len(updn),(',').join(updn))

        repeated_coord = set([x for x in lista if coord.count(x[1:]) > 1])

        non_repeated_coord = set([x for x in lista if coord.count(x[1:]) == 1])

        numbers = []

        #if  in coords hay numeros, etonces error si any es > dim
        
        ErrorIndexAlphaNum = False
        ErrorIndexInt = False

        for i in coord:
            
            if (not i.isalnum()) or ("." in i):
                
                ErrorIndexAlphaNum= True
                
            try:
                
                if (int(i) >= dim) or (int(i) < 0):
                    
                    ErrorIndexInt = True
                    
            except:
                
                pass

            if ErrorIndexAlphaNum:
                
                raise SyntaxError("Wrong indices. Every index must be a name or a integer without any special characters.")
            
            elif ErrorIndexInt:
                
                raise SyntaxError("Wrong indices. Every index must be a name or a integer greater than 0 and less than the dimension.")
                


        for i in coord:

            if i in np.asarray([range(dim)],dtype=str):
                
                numbers.append(i)

        # if numbers esta bien, luego tiene que ir un elif len(numbers) != 0. En ese caso hay problemas
        # finalmente tiene que ir otro elif, que es si no tiene numeros. Este caso se dividira en si tiene o no indices repetidos.
                
        if len(numbers) == len(coord): # All numbered indices
            
            return self.__scalar_call(updn,numbers,Nindex,ten_call)

            
        if len(numbers) != 0: # Some numbered indices and letters


            ## Si hay numeros no suma indices repetidos
            ## En este caso se podría tener que retorne escalar, lo que puede terminar en error.
            
            return self.__Tdata_call(updn,coord,numbers,Nindex,dim,ten_call)

        for i in lista:

            if lista.count(i) > 1: 

                raise SyntaxError('Problem with the indices. Error in the Einstein summation.')


        if len(repeated_coord) == 0:

            elements = eval('self.%s[Nindex]'%ten_call,locals(),globals())

            return Tdata(str_index,elements)

        else: # Repeated Index

            new_string = ''

            old_string = ''

            return_string = ''

            new_rank = len(lista) - 2

            if new_rank != 0:

                temp = construct(0,dim,new_rank)

            k = 0
            rep_count = 0
            variable = ''

            for var in lista:

                if coord.count(var[1:]) == 2 and rep_count != 2:

                    if variable == var[1:] or variable == '':

                        old_string += '[q]'

                        rep_count += 1

                        variable = var[1:]

                    else:# coord.count(var[1:]) == 1 :

                        old_string += '[p[%d]]'%k

                        return_string += '%s,'%var

                        k += 1


                else:# coord.count(var[1:]) == 1 :

                    old_string += '[p[%d]]'%k

                    return_string += '%s,'%var

                    k += 1

            return_string = return_string[:-1]

            for k in range(new_rank):

                new_string += '[p[%d]]'%k


            if new_rank != 0:

                for p in iterprod(range(dim), repeat = new_rank):

                    var_temp = 0

                    for q in range(dim):

                        var_temp += eval('self.%s[Nindex]%s'%(ten_call,old_string),locals(),globals())

                        exec('temp%s = var_temp'%new_string,locals(),globals())

                data_result = Tdata(return_string,temp)

                TEMP = Tensor('TEMP',new_rank)

                TEMP.assign(data_result,return_string,printing=False)

                return TEMP(return_string)

            else:
                
                var = 0

                for q in range(dim):

                    var += eval('self.%s[Nindex]%s'%(ten_call,old_string),locals(),globals())

                return var 

    def __assign(self, elements, index=None,All = False,printing = False):

        '''
        # Revisar el nombre de printing. Puede ser Verbose

        It assigns the elements to the tensor on the corresponding index. 
        If All = True, then it computes the thensor with the rest of the indices combinations.

        index = '^,^,_'
        elements = [[[0,1,2,3],[4,5,6,7],[8,9,10,11],[12,13,14,15]]]

        NOTE: the argument "elements" has to be shaped like the example so the indexation goes like elements[i][j][k]

        '''
        dim = config.dim

        if isinstance(elements,Tdata):

            new_lista = index.split(',')

            old_lista = elements.full_index.split(',')

            rank = len(old_lista)

            New_data = construct(0,dim,rank)

            new_index = ''
            old_index = ''

            for i in range(rank):

                new_index += '[p[%d]]'%i

            for j in range(rank):

                i = 0

                while old_lista[j][1:] != new_lista[i][1:]:

                    i += 1

                old_index += '[p[%d]]'%i

            for p in iterprod(range(dim),repeat=rank):

                string = 'New_data%s = elements.elements%s'%(new_index,old_index)

                exec(string,locals(),globals())

            new_updn = (',').join(x[0] for x in new_lista)

            self.assign(New_data,new_updn,printing=False)

        elif index == None:

            print(ERROR)

        else:

            dim = config.dim

            k = compare(self.n,index) # numero correspondiente a '^,^' o '_,^', etc


            for p in iterprod(range(dim),repeat=self.n):
                            
                string = 'self.tensor[%s]'%k 
                string2 = ''

                for l in p:
                        
                    string2 += '[%s]'%l

                string = '%s%s = elements%s'%(string,string2,string2)

                try: 

                    exec(string,locals(),globals())

                except AttributeError:

                    pass
            
            self.indices[k] = True

            if printing == True:
        
                print('Elements assigned correctly to the components %s'%index)

            if All == True:

                self.indexcomb(index)

        self.space()

    def space(self):
        
        '''
        Retorna solo las componentes espaciales del tensor completo
        
        Generalizar para todos los indices. El [1:,1:,....,1:]
        
        
        '''
        
        for k in range(2**self.n):
            
            index = k
        
            iterstring = ''
            iterstring2 = ''

            for i in range(self.n):

                iterstring += '[p[%d]]'%i
                iterstring2 += '[p[%d]+1]'%i

            for p in iterprod(range(config.dim-1),repeat=self.n):

                exec_str = 'self.tensor_sp[%d]%s = self.tensor[%d]%s'%(index,iterstring,index,iterstring2)

                exec(exec_str,locals(),globals())

    def display(self, index=None, aslist = None, simplify = False, spatial = None):

        if config.space_time == False and spatial is None:

            spatial = True

        if spatial == True:

            return self.display_spatial(index, aslist, simplify)

        if core_calc == 'sp' and simplify == True:

            warn("The simplify argument is intended to be used only with giacpy.\n The result is not affected when using Sympy.")
        
        if index is None:

            index = self.sequence[0]

        if aslist == None:

            if self.n <= 2:

                aslist = False
            
            else:

                aslist = True

        dim = config.dim
        
        k = 0
        for i in self.sequence:
            if i == index:
                break 
            k += 1

        if core_calc == 'sp':

            init_printing()

        if k == len(self.sequence):

            raise ValueError('Bad index definition')

        if index == '' and self.n == 0: # Scalar

            display(self.tensor)
        
        
        elif aslist == False:

            # if k == len(self.sequence):

            #     raise ValueError('Bad index definition')

            if self.n == 1 and index == '^':
                
                if core_calc == 'sp':

                    display_IP(Array(self.tensor[k]).reshape(dim,1))

                elif core_calc == 'gp':

                    f = io.StringIO()

                    with redirect_stdout(f):

                        print(latex(giac(self.tensor[k]).transpose()))
                    out = f.getvalue()

                    out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                    display_IP(Math_IP(gp_pretty_latex(out)))

            else:

                if core_calc == 'sp':
            
                    display_IP(Array(self.tensor[k]))

                elif core_calc == 'gp':

                    if simplify == False:

                        f = io.StringIO()

                        if self.n != 1:
                            with redirect_stdout(f):
                                print(latex(matrix(self.tensor[k])))
                        else:
                             with redirect_stdout(f):
                                print(latex(giac(self.tensor[k])))
                        out = f.getvalue()

                        out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                        display_IP(Math_IP(gp_pretty_latex(out)))

                    else:

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print((self.tensor[k]))
                        out = f.getvalue()

                        string = sp_latex(sp_simplify(sp_Array(sp_sympify(out))))

                        display_IP(Math_IP(string))

            
        else:

            count = 0
                
            for p in iterprod(range(dim),repeat=self.n):
                    
                string = 'valor = self.tensor[%s]'%k

                for l in p:
                        
                    string += '[%s]'%l

                exec(string,locals(),globals())
                
                str_name =  "\\varepsilon"
                
                string = "{%s}"%str_name
                    
                i = 0
                    
                for l in index.split(','):
                        
                    if core_calc == 'gp':

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print(p[i])
                        out = f.getvalue()
                            
                        string = '%s{}%s{%s}'%(string,l,out)
                    
                    elif core_calc == 'sp':

                        string = '%s{}%s{%s}'%(string,l,str(p[i]))

                    i += 1
                    
                if valor != 0:
                        
                    #string += " = %s"% (latex(valor))
                    
                    #display_IP(Math(string))

                    if core_calc == 'gp':

                        if simplify == False:

                            display(valor,string)

                        else:

                            f = io.StringIO()
                            with redirect_stdout(f):
                                print(valor)
                            out = f.getvalue()

                            string2 = sp_latex(sp_simplify(sp_sympify(out)))

                            string = "%s = %s"%(string,string2)

                            display_IP(Math_IP(string))

                    elif core_calc == 'sp':

                        display(valor,string)

                    count += 1

            if count == 0:

                print('All components are zero')

    def display_spatial(self, index=None, aslist = None, simplify = False):

        if core_calc == 'sp' and simplify == True:

            warn("The simplify argument is intended to be used only with giacpy.\n The result is not affected when using Sympy.")
        
        if index is None:

            index = self.sequence[0]

        rank = self.n

        if aslist == None:

            if rank <= 2:

                aslist = False
            
            else:

                aslist = True

        dim = config.dim - 1 # Se elimina la dimension temporal
        
        k = 0
        for i in self.sequence:
            if i == index:
                break 
            k += 1

        if core_calc == 'sp':

            init_printing()

        if k == len(self.sequence):

            raise ValueError('Bad index definition')

        if index == '' and rank == 0: # Scalar

            display(self.tensor_sp)
        
        
        elif aslist == False:

            # if k == len(self.sequence):

            #     raise ValueError('Bad index definition')

            if rank == 1 and index == '^':
                
                if core_calc == 'sp':

                    display_IP(Array(self.tensor_sp[k]).reshape(dim,1))

                elif core_calc == 'gp':

                    f = io.StringIO()

                    with redirect_stdout(f):

                        print(latex(giac(self.tensor_sp[k]).transpose()))
                    out = f.getvalue()

                    out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                    display_IP(Math_IP(gp_pretty_latex(out)))

            else:

                if core_calc == 'sp':
            
                    display_IP(Array(self.tensor_sp[k]))

                elif core_calc == 'gp':

                    if simplify == False:

                        f = io.StringIO()

                        if rank != 1:
                            with redirect_stdout(f):
                                print(latex(matrix(self.tensor_sp[k])))
                        else:
                             with redirect_stdout(f):
                                print(latex(giac(self.tensor_sp[k])))
                        out = f.getvalue()

                        out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                        display_IP(Math_IP(gp_pretty_latex(out)))

                    else:

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print((self.tensor_sp[k]))
                        out = f.getvalue()

                        string = sp_latex(sp_simplify(sp_Array(sp_sympify(out))))

                        display_IP(Math_IP(string))

            
        else:

            count = 0
                
            for p in iterprod(range(dim),repeat=rank):
                    
                string = 'valor = self.tensor_sp[%s]'%k

                for l in p:
                        
                    string += '[%s]'%l

                exec(string,locals(),globals())
                
                str_name =  "\\varepsilon"
                
                string = "{%s}"%str_name
                    
                i = 0
                    
                for l in index.split(','):
                        
                    if core_calc == 'gp':

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print(p[i])
                        out = f.getvalue()
                            
                        string = '%s{}%s{%s}'%(string,l,out)
                    
                    elif core_calc == 'sp':

                        string = '%s{}%s{%s}'%(string,l,str(p[i]))

                    i += 1
                    
                if valor != 0:
                        
                    #string += " = %s"% (latex(valor))
                    
                    #display_IP(Math(string))

                    if core_calc == 'gp':

                        if simplify == False:

                            display(valor,string)

                        else:

                            f = io.StringIO()
                            with redirect_stdout(f):
                                print(valor)
                            out = f.getvalue()

                            string2 = sp_latex(sp_simplify(sp_sympify(out)))

                            string = "%s = %s"%(string,string2)

                            display_IP(Math_IP(string))

                    elif core_calc == 'sp':

                        display(valor,string)

                    count += 1

            if count == 0:

                print('All components are zero')


import numpy as np 
from itertools import product as iterprod
from numpy.lib.arraysetops import isin
from tqdm import tqdm_notebook
from warnings import warn
from .core import config
from .core.core import *
from .core.display import display
from .core.interp import syntax, add_examine, mul_examine, der_examine
from .core.tdata import Tdata, construct

if core_calc == 'gp':

    import io
    from contextlib import redirect_stdout
    from .core.display import gp_pretty_latex
    from .core.core import sp_simplify, sp_sympify, sp_latex, sp_Array

def reload_all(new_module):

    ''' x

    This function reloads all the variables in __all__ from module
        
    Thanks to hurturk on https://stackoverflow.com/questions/44492803/python-dynamic-import-how-to-import-from-module-name-from-variable?newreg=9312458d429647fc8ec2a72e5655d197
        
    '''

    string = "globals().update({{n: getattr(module, n) for n in module.__all__}} if hasattr(module, '_all_') else {k: v for (k, v) in module.__dict__.items() if not k.startswith('_')})"

    string = string.replace('module',new_module)

    return string

def ordenar(n):

    '''
    Returns a list with all the possible combinations of indices for a tensor of n indices [['_','_'],['^','_'],['_','^'],['^','^']]
    '''

    return [','.join(i) for i in iterprod(['_','^'],repeat = n)]


class gError(Exception):

    pass

def compare(n,string):

    '''
    Returns de index corresponding to the element in the list of all the combination of indices that mathces string
    '''

    if n == 0:

        return None

    lista = ordenar(n)

    if string not in lista:

        raise ValueError('Bad index definition') # Verificar que se levante el error correcto

    for i in range(len(lista)):

        if lista[i] == string:

            return i # esto es un numero

        else:
            pass


def bajarindice(tensor,i,kstring,kstring2):

    # i: primera posicion que tiene un indice arriba ej: en un string de indices '_,^,^' -> i = 1 (contando desde 0)

    # kstring: string original de los indices tensor ej: '_,^,^'

    # kstring2: string objetivo (con el indice i abajo) de los indices del tencor ej: '_,_,^'

    index = compare(tensor.n,kstring)

    index2 = compare(tensor.n,kstring2)

    dim = config.dim

    greek = config.greek

    g = config.g.tensor

    NAME = tensor.name

    if NAME in config.default_tensors:

        NAME = config.default_tensors[NAME]

    #---

    iterstring = ''

    greek_desc = ''

    iterg = ''

    count2 = 0

    for count in range(tensor.n):

        if count == i:

            iterstring += '[j]'

            iterg += '[p[%d]][j]'%i

            count2 += 1

        else:

            iterstring += '[p[%d]]'%count2

            count2 += 1

    iterstring2 = ''

    for count in range(tensor.n):

        greek_desc += '{%s%s}'%(kstring2.split(',')[count],greek[count])

        iterstring2 += '[p[%d]]'%count

    #---

    description =  '%s Tensor $%s%s$'%(tensor.name,NAME,greek_desc)

    for p in tqdm_notebook(iterprod(range(dim),repeat=tensor.n),total=dim**tensor.n,desc= description):

        temporal = 0

        for j in range(dim):

            bla2 = 'g[0]%s*tensor.tensor[index]%s'%(iterg,iterstring)

            temporal += eval(bla2)

        if config.ord_status == True:

            bla = 'tensor.tensor[index2]%s = tensor_series(temporal)'%iterstring2 

        else:

            bla = 'tensor.tensor[index2]%s = temporal'%iterstring2       # Esto asigna a los indices que queremos llegar

        exec(bla,globals(),locals())

    tensor.indices[index2] = True

    # hasta aqui todo okidoki uwu

    # ESTA FUNCION SE TENDRA QUE LLAMAR CUANTAS VECES SEA NECESARIA HASTA LLEGAR A TENER TODO ABAJO _ _ _

    # REVISAR EL RESULTADO DE LA PRIMERA BAJADA DE INDICES COMPARAR CON GRTENSOR



def subirindice(tensor,kstring):

    # Recibe:

    # 1) Nombre de la variable de clase Tensor al cual queremos subir los indices

    # 2) string de la forma '_,^,_'

    index = compare(tensor.n,kstring)

    klista = kstring.split(',') # ['_','^','_']

    NAME = tensor.name

    dim = config.dim

    greek = config.greek

    g = config.g.tensor

    if NAME in config.default_tensors:

        NAME = config.default_tensors[NAME]

    for i in range(len(klista)): # _,_

        klista2 = list(klista)

        if klista[i] == '_': # _,^ 

            klista2[i] = '^'

            kstring2 = ','.join(klista2)

            index2 = compare(tensor.n,kstring2) # numero correspondiente a '^,^' o '^,_' del tensor

            if tensor.indices[index2] == True: # si ya esta calculado no hace nada

                pass

            else:

                bla = ''

                iterstring = ''

                iterstring2 = ''

                greek_desc = ''

                iterg = ''

                cosa2 = 0

                for count in range(tensor.n):

                    greek_desc += '{%s%s}'%(klista2[count],greek[count])

                    iterstring2 += '[p[%d]]'%count

                    if count == i:

                        iterstring += '[j]'

                        iterg += '[p[%d]][j]'%i

                        cosa2 += 1

                    else:

                        iterstring += '[p[%d]]'%cosa2

                        cosa2 += 1

                description =  '%s Tensor $%s%s$'%(tensor.name,NAME,greek_desc)

                #---

                for p in tqdm_notebook(iterprod(range(dim),repeat=tensor.n),total=dim**tensor.n,desc= description):

                    temporal = 0

                    for j in range(dim):

                        bla2 = 'g[3]%s*tensor.tensor[index]%s'%(iterg,iterstring)

                        temporal += eval(bla2)

                    if config.ord_status == True:

                        bla = 'tensor.tensor[index2]%s = tensor_series(temporal)'%iterstring2 

                    else:

                        bla = 'tensor.tensor[index2]%s = temporal'%iterstring2      # Esto asigna a los indices que queremos llegar

                    #print('En subir indice\t', iterg, iterstring, iterstring2)

                    exec(bla,globals(),locals())

                    tensor.indices[index2] = True

            subirindice(tensor,kstring2)

def createfirstindex(tensor,kstring):

    # kstring: string original de los indices ej: '^,_,_' ...
    # tensor: tensor de la clase Tensor

    N = tensor.n # Number of index

    objetivo = tensor.sequence[0].split(',') # ['_','_','_']

    lista = kstring.split(',')  # ['^','_','_']

    for i in range(len(lista)):

        if lista[i] != '_':

            lista[i] = '_'

            kstring_new = ','.join(lista)

            bajarindice(tensor,i,kstring,kstring_new)

            kstring = kstring_new

class Tensor:

    r""" Tensor Class

    The 'Tensor' class contains all the relevant information for a Tensor object. 

    Attributes
    ----------
    name : str
        name of the Tensor.
    n : int
        rank of the Tensor.
    dim :   int
        dimension of the Tensor.
    sequence : list
        a list containing all the possible combination of indices for the Tensor.
    indices : array_like
        an array containing Boolean values to indicate which combination of the indices were calculated.
    tensor : array_like
        an array containing each component of the tensor. Initialized with NaN values.
    tensor_sp : array_like
        an array containing only the spatial components of the tensor. Initialized with NaN values.
    
    Methods
    -------
    assign(elements,index=None, All = False, printing = True, spatial = None)
        It assigns the elements to the tensor on the corresponding indices combination. 
    complete(index)
        It calculates the missing indices combintation of the tensor raising and lowering the indices from the given indices combination.
    display(index=None, aslist = None, simplify = False, spatial=None)
        Displays the Tensor. By default it displays the tensor with all covariant indices unless index is given.
    series(index = None)
        Performs a series expanson on each component of the Tensor. If index is given, only acts on the given indices combination.
    simplify(index = None)
        Simplifies each component of the tensor. If index is given, only acts on the given indices combination.
    expand(index = None)
        Expands each component of the tensor. If index is given, only acts on the given indices combination.
    factor(index = None)
        Factorizes each component of the tensor. If index is given, only acts on the given indices combination.

    Raises
    ------
    ValueError
        If the rank of a tensor is not a positive integer. If the Dimension is undefined
    """
    
    
    

    def __init__(self,name,n):

        if n <= 0 or not isinstance(n,int):
            
            raise ValueError("The rank of a tensor (n) must be a positive integer.")

        dim = config.dim

        if dim == 0:

            raise ValueError('Dimension undefined')

        if name == 'g':

            if config.g == True:

                while True:

                    answer = input('Warning g is already defined as the metric tensor. Are you sure that you want to overwrite it?\nyes/no?')

                    if answer == 'yes' or answer == 'y' or answer == 'Yes':

                        print('g has been overwritten')

                        break

                    if answer == 'no' or answer == 'n' or answer == 'No':

                        raise gError

                    else:

                        print('Please try again.\n')

        self.name = name

        self.n = n

        self.sequence = ordenar(n)

        self.indices = np.full((2**n), False)

        if core_calc == 'sp':

            string = 'sympify(np.nan),'*dim

        elif core_calc == 'gp':
            
            string = 'np.nan,'*dim

        string = '['+string[:-1]+'],'

        for k in range(n-1):

            string = '['+(string*dim)[:-1]+'],'  

        string = 'self.tensor = [' + (string*(2**n))[:-1]+']'

        exec(string)

        if core_calc == 'sp':

            string = 'sympify(np.nan),'*(dim-1)

        elif core_calc == 'gp':
            
            string = 'np.nan,'*(dim-1)

        string = '['+string[:-1]+'],'

        for k in range(n-1):

            string = '['+(string*(dim-1))[:-1]+'],'  

        string = 'self.tensor_sp = [' + (string*(2**n))[:-1]+']'

        exec(string)

    def __repr__(self):

        return '%s tensor defined'%self.name

    def __str__(self):

        if self.n <=2:

            self.display()

            string = ''

        else:

            string = 'Too many tensor indices. Please use the display method.'
        
        return string

    def __scalar_call(self,updn,numbers,Nindex,ten_call):
    
        tdat_str = ''
        
        new_string_right = ''
        
        for i in numbers:
                
            new_string_right += '[%s]'%i
        
        return eval('self.%s[%s]%s'%(ten_call,Nindex,new_string_right))

    def __Tdata_call(self,updn,coord,numbers,Nindex,dim,ten_call):
    
        tdat_str = ''
        
        new_string_right = ''
        
        k = 0
        j = 0
        
        for i in range(len(updn)):
            
            if coord[i] not in numbers:
            
                tdat_str += updn[i] + coord[i] + ','
                
                new_string_right += '[p[%d]]'%k
                
                k += 1
                
            else: 
                
                new_string_right += '[%s]'%numbers[j]
                
                j += 1

        tdat_str = tdat_str[:-1]
                
                    
        new_rank = len(coord)-len(numbers)
        
        elements = construct(0,dim,new_rank)
        
        new_string_left = ''
        
        for k in range(new_rank):

            new_string_left += '[p[%d]]'%k
        
        
        for p in iterprod(range(dim), repeat = new_rank):

            exec_str = 'elements%s = self.%s[%s]%s'%(new_string_left,ten_call,Nindex,new_string_right)
        
            exec(exec_str,locals(),globals())

        data_result = Tdata(tdat_str,elements)

        TEMP = Tensor('TEMP',new_rank)

        TEMP.assign(data_result,tdat_str,printing=False)

        return TEMP(tdat_str)

    def __call__(self,str_index):

        '''Call Method for Tensor Class.

        Returns a Tdata object with the indices given as argument of the call.
        
        '''
        # Solo acepta a lo mas la contraccion de 1 solo indice. Hay que incluir esto en el examine de este call

        syntax(str_index,self.n)

        if config.space_time == 1:

            dim = config.dim

            ten_call = 'tensor'

        else:

            dim = config.dim - 1

            ten_call = 'tensor_sp'

        lista = str_index.split(',')

        updn = [symbol[0] for symbol in lista]

        coord = [symbol[1:] for symbol in lista]
        
        Nindex = compare(len(updn),(',').join(updn))

        repeated_coord = set([x for x in lista if coord.count(x[1:]) > 1])

        non_repeated_coord = set([x for x in lista if coord.count(x[1:]) == 1])

        numbers = []

        #if  in coords hay numeros, etonces error si any es > dim
        
        ErrorIndexAlphaNum = False
        ErrorIndexInt = False

        for i in coord:
            
            if (not i.isalnum()) or ("." in i):
                
                ErrorIndexAlphaNum= True
                
            try:
                
                if (int(i) >= dim) or (int(i) < 0):
                    
                    ErrorIndexInt = True
                    
            except:
                
                pass

            if ErrorIndexAlphaNum:
                
                raise SyntaxError("Wrong indices. Every index must be a name or a integer without any special characters.")
            
            elif ErrorIndexInt:
                
                raise SyntaxError("Wrong indices. Every index must be a name or a integer greater than 0 and less than the dimension.")
                


        for i in coord:

            if i in np.asarray([range(dim)],dtype=str):
                
                numbers.append(i)

        # if numbers esta bien, luego tiene que ir un elif len(numbers) != 0. En ese caso hay problemas
        # finalmente tiene que ir otro elif, que es si no tiene numeros. Este caso se dividira en si tiene o no indices repetidos.
                
        if len(numbers) == len(coord): # All numbered indices
            
            return self.__scalar_call(updn,numbers,Nindex,ten_call)

            
        if len(numbers) != 0: # Some numbered indices and letters


            ## Si hay numeros no suma indices repetidos
            ## En este caso se podría tener que retorne escalar, lo que puede terminar en error.
            
            return self.__Tdata_call(updn,coord,numbers,Nindex,dim,ten_call)

        for i in lista:

            if lista.count(i) > 1: 

                raise SyntaxError('Problem with the indices. Error in the Einstein summation.')


        if len(repeated_coord) == 0:

            elements = eval('self.%s[Nindex]'%ten_call,locals(),globals())

            return Tdata(str_index,elements)

        else: # Repeated Index

            new_string = ''

            old_string = ''

            return_string = ''

            new_rank = len(lista) - 2

            if new_rank != 0:

                temp = construct(0,dim,new_rank)

            k = 0
            rep_count = 0
            variable = ''

            for var in lista:

                if coord.count(var[1:]) == 2 and rep_count != 2:

                    if variable == var[1:] or variable == '':

                        old_string += '[q]'

                        rep_count += 1

                        variable = var[1:]

                    else:# coord.count(var[1:]) == 1 :

                        old_string += '[p[%d]]'%k

                        return_string += '%s,'%var

                        k += 1


                else:# coord.count(var[1:]) == 1 :

                    old_string += '[p[%d]]'%k

                    return_string += '%s,'%var

                    k += 1

            return_string = return_string[:-1]

            for k in range(new_rank):

                new_string += '[p[%d]]'%k


            if new_rank != 0:

                for p in iterprod(range(dim), repeat = new_rank):

                    var_temp = 0

                    for q in range(dim):

                        var_temp += eval('self.%s[Nindex]%s'%(ten_call,old_string),locals(),globals())

                        exec('temp%s = var_temp'%new_string,locals(),globals())

                data_result = Tdata(return_string,temp)

                TEMP = Tensor('TEMP',new_rank)

                TEMP.assign(data_result,return_string,printing=False)

                return TEMP(return_string)

            else:
                
                var = 0

                for q in range(dim):

                    var += eval('self.%s[Nindex]%s'%(ten_call,old_string),locals(),globals())

                return var 

    def space(self):
        
        '''
        Saves the spatial components of the Tensor on the attribut tensor_sp.

        '''
        
        for k in range(2**self.n):
            
            index = k
        
            iterstring = ''
            iterstring2 = ''

            for i in range(self.n):

                iterstring += '[p[%d]]'%i
                iterstring2 += '[p[%d]+1]'%i

            for p in iterprod(range(config.dim-1),repeat=self.n):

                exec_str = 'self.tensor_sp[%d]%s = self.tensor[%d]%s'%(index,iterstring,index,iterstring2)

                exec(exec_str,locals(),globals())

    def series(self,index = None):

        '''
        Expands each element of the tensor at the given indices.
        
        '''

        if index is None:

            index = self.sequence[0]

        dim = config.dim

        k = 0
        for i in self.sequence:
            if i == index:
                break 
            k += 1

        iterstring = ""

        for i in range(self.n):

            iterstring += '[p[%d]]'%i

        for p in iterprod(range(dim),repeat=self.n):

            execstr = "self.tensor[k]%s = tensor_series(self.tensor[k]%s)"%(iterstring,iterstring)

            exec(execstr,locals(),globals())
   
    def complete(self, index):

        '''Calculates the missing indices combintation of the tensor

        Calculates the missing indices combintation of the tensor using the metric to raise and lower the indices, starting from the given indices combination

        Parameters
        ----------
        index : str, optional
            string indicating the indices combination used to start the complete method

        Raises
        ------
        ValueError
            If the 'index' specified does not correspond to any possible indices combination.
        - string indicating the starting indices combination i.e., '_,^,_' for a 3-rank tensor.
        
        '''

        if (index not in self.sequence and index is not None) or index == '':

            raise ValueError('Bad index definition')

        createfirstindex(self,index) 

        lista = []

        for i in range(self.n):

            lista.append('_,')

        string = ''.join(lista)

        string = string[:-1] 

        subirindice(self,string)

        NAME = self.name 

        if NAME in config.default_tensors:

            NAME = config.default_tensors[NAME]

        display_string = 'All other indices of %s Tensor $%s$  already calculated.'%(self.name,NAME)

        self.space()

        display_IP(Latex_IP(display_string))

    def __assign_space(self,elements,index,printing = True):

        # Revisar el nombre de printing. Puede ser Verbose

        dim = config.dim - 1

        if isinstance(elements,Tdata):

            new_lista = index.split(',')

            old_lista = elements.full_index.split(',')

            rank = len(old_lista)

            New_data = construct(0,dim,rank)

            new_index = ''
            old_index = ''

            for i in range(rank):

                new_index += '[p[%d]]'%i

            for j in range(rank):

                i = 0

                while old_lista[j][1:] != new_lista[i][1:]:

                    i += 1

                old_index += '[p[%d]]'%i

            for p in iterprod(range(dim),repeat=rank):

                string = 'New_data%s = elements.elements%s'%(new_index,old_index)

                exec(string,locals(),globals())

            new_updn = (',').join(x[0] for x in new_lista)

            self.__assign_space(New_data,new_updn,printing=False)

        elif index == None:

            raise ValueError("'index' must be a string. e.g.: '^,^,_,^'. ")

        else:

            dim = config.dim - 1

            k = compare(self.n,index) # numero correspondiente a '^,^' o '_,^', etc


            for p in iterprod(range(dim),repeat=self.n):
                            
                string = 'self.tensor_sp[%s]'%k 
                string2 = ''

                for l in p:
                        
                    string2 += '[%s]'%l

                string = '%s%s = elements%s'%(string,string2,string2)

                try: 

                    exec(string,locals(),globals())

                except AttributeError:

                    pass
            
            self.indices[k] = True

            if printing == True:
        
                print('Elements assigned correctly to the components %s'%index)

    def assign(self, elements, index=None, All = False, printing = True, spatial = None):

        """It assigns the elements to the tensor on the corresponding indices combination.

        Simplifies each component of the tensor. If index is given, only acts on the given indices combination.

        Parameters
        ----------
        elements : array_like or Tdata
            elements to be assigned to the Tensor indices combination. These elemenents must be given in an array-like form such as a numpy array or a nested list.
            Otherwise, elements can be a 'Tdata' object obtained through the call method of a 'Tensor' object.
        index : str, optional
            string indicating the Tensor indices where to assign the elements.
        All : boolean, optional
            boolean indicating if the rest of indices combinations should be calculated after assigning the 'elements' to the Tensor 
            (the default value is False, it implies that no other indices combination will be calculated).
        printing : boolean, optional
            boolean indicating if the method should print the success after assigning the 'elements' to the Tensor
            (the default value is True, it implies that the method will indicate on which indices combination the 'elements' have been assigned).
        spatial : boolean, optional
            boolean indicating if the assignation of the 'elements' corresponds only to the spatial components of the Tensor.
            (the default value is None, it checks if the user is working only with space coordinates and if that's the case, only assigns the spatial components, else, it assigns the 'elements' to the full Tensor).
            
        Raises
        ------
        ValueError
            #####If the 'index' specified does not correspond to any possible indices combination.
        
        Examples
        --------
        Assigning a nested list to a 'Tensor'

        >>> elements = [[-1/a**2, 0, 0, 0], [0, 1, 0, 0], [0, 0, r**2, 0], [0, 0, 0, r**2*sin(theta)**2]] # the 'sin' function is from sympy
        >>> A = pt.ten('A', 2)
        >>> A.assign(elements, '_i,_j')
        
        Elements assigned correctly to the _i,_j components


        Assigning a 'tdata' to a 'Tensor

        >>> elements = g('_i,_j')/a**2 # using the metric already defined
        >>> A = pt.ten('A',2)
        >>> A.assign(elements,'_i,_j')

        Elements assigned correctly to the _i,_j components

        """

        if config.space_time == False and spatial is None:

            spatial = True

        if spatial == True:

            return self.__assign_space(elements,index,printing)

        dim = config.dim

        if isinstance(elements,Tdata):

            new_lista = index.split(',')

            old_lista = elements.full_index.split(',')

            rank = len(old_lista)

            New_data = construct(0,dim,rank)

            new_index = ''
            old_index = ''

            for i in range(rank):

                new_index += '[p[%d]]'%i

            for j in range(rank):

                i = 0

                while old_lista[j][1:] != new_lista[i][1:]:

                    i += 1

                old_index += '[p[%d]]'%i

            for p in iterprod(range(dim),repeat=rank):

                string = 'New_data%s = elements.elements%s'%(new_index,old_index)

                exec(string,locals(),globals())

            #new_updn = (',').join(x[0] for x in new_lista)

            self.assign(New_data, index, All, printing)

        elif index == None:

            raise ValueError("'index' must be a string. e.g.: '^i,^j,_k,^l'. ")

        else:

            dim = config.dim

            updn = (',').join(x[0] for x in index.split(','))

            k = compare(self.n,updn) # numero correspondiente a '^,^' o '_,^', etc


            for p in iterprod(range(dim),repeat=self.n):
                            
                string = 'self.tensor[%s]'%k 
                string2 = ''

                for l in p:
                        
                    string2 += '[%s]'%l

                string = '%s%s = elements%s'%(string,string2,string2)

                try: 

                    exec(string,locals(),globals())

                except AttributeError:

                    pass
            
            self.indices[k] = True

            if printing == True:
        
                print('Elements assigned correctly to the %s components'%index)

            if All == True:

                self.complete(updn)

        self.space()

    def simplify(self,index=None):

        """Simplifies each component of the tensor.

        Simplifies each component of the tensor. If index is given, only acts on the given indices combination.

        Parameters
        ----------
        index : str, optional
            string indicating the indices combination to apply the simplification (the default is None, it implies that it will simplify all the inices combinations of the Tensor).

        Raises
        ------
        ValueError
            If the 'index' specified does not correspond to any possible indices combination.
        """

        dim = config.dim

        if (index not in self.sequence and index is not None) or index == '':

            raise ValueError('Bad index definition')

        if config.space_time == 1:

            dim = config.dim

            ten_call = 'tensor'

        else:

            dim = config.dim - 1

            ten_call = 'tensor_sp'
        

        if index is None:

            listax = list(range(2**self.n))

            if self.n == 0:


                self.tensor = eval("self.%s.simplify()"%ten_call,locals(),globals())

        else:

            listax = [compare(self.n,index)]
        
        for k in listax: 
            
            if self.indices[k] == True:

                if self.name == 'Riemann' and k == 0: 

                    Riemann_list = construct('False',dim,4)

                    for p in iterprod(range(dim),repeat=self.n):

                        counta = p[0]
                        countb = p[1]
                        countc = p[2]
                        countd = p[3]

                        if Riemann_list[counta][countb][countc][countd] == False:
                            
                            string = 'self.%s[%s]'%(ten_call,k) 

                            for l in p:
                                    
                                string += '[%s]'%l

                            string = string + '=' + string + '.simplify()'

                            exec(string,locals(),globals())

                            print('Fully Simplified:  ',counta,countb,countc,countd)

                            # Skew Symmetry

                            if Riemann_list[counta][countb][countd][countc] == False:

                                self.tensor[0][counta][countb][countd][countc] = -self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][counta][countb][countd][countc] = -self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[counta][countb][countd][countc] = True

                            if Riemann_list[countb][counta][countc][countd] == False:

                                self.tensor[0][countb][counta][countc][countd] = -self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][countb][counta][countc][countd] = -self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[countb][counta][countc][countd] = True

                            # Interchange Symmetry

                            if Riemann_list[countc][countd][counta][countb] == False:

                                self.tensor[0][countc][countd][counta][countb] = self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][countc][countd][counta][countb] = self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[countc][countd][counta][countb] = True

                            # Bianchi Identity (First)

                            if Riemann_list[counta][countc][countd][countb] == False and Riemann_list[counta][countd][countb][countc] == True:

                                self.tensor[0][counta][countc][countd][countb] = -self.tensor[0][counta][countb][countc][countd] - self.tensor[0][counta][countd][countb][countc]
                                self.tensor_sp[0][counta][countc][countd][countb] = -self.tensor_sp[0][counta][countb][countc][countd] - self.tensor_sp[0][counta][countd][countb][countc]

                                Riemann_list[counta][countc][countd][countb] = True

                            if Riemann_list[counta][countd][countb][countc] == False and Riemann_list[counta][countc][countd][countb] == True:

                                self.tensor[0][counta][countd][countb][countc] = -self.tensor[0][counta][countb][countc][countd] - self.tensor[0][counta][countc][countd][countb]
                                self.tensor_sp[0][counta][countd][countb][countc] = -self.tensor_sp[0][counta][countb][countc][countd] - self.tensor_sp[0][counta][countc][countd][countb]

                                Riemann_list[counta][countd][countb][countc] = True


                            Riemann_list[counta][countb][countc][countd] = True
                    
                else:
                
                    for p in iterprod(range(dim),repeat=self.n):
                                
                        string = 'self.%s[%s]'%(ten_call,k) 

                        for l in p:
                                
                            string += '[%s]'%l

                        if config.ord_status == False:

                            string = string + '= simplify(' + string + ')'

                        else:

                            string = string + '= simplify(tensor_series(' + string + '))'

                        #print(string)

                        exec(string,locals(),globals())

        if config.space_time == 1:
        
            self.space()

    def expand(self,index=None):

        """Expands each component of the tensor.

        Expands each component of the tensor. If index is given, only acts on the given indices combination.

        Parameters
        ----------
        index : str, optional
            string indicating the indices combination to apply the expansion (the default is None, it implies that it will expand all the inices combinations of the Tensor).

        Raises
        ------
        ValueError
            If the 'index' specified does not correspond to any possible indices combination.
        """

        dim = config.dim

        if (index not in self.sequence and index is not None) or index == '':

            raise ValueError('Bad index definition')

        
        if config.space_time == 1:

            dim = config.dim

            ten_call = 'tensor'

        else:

            dim = config.dim - 1

            ten_call = 'tensor_sp'
        

        if index is None:

            listax = list(range(2**self.n))

            if self.n == 0:


                self.tensor = eval("self.%s.expand()"%ten_call,locals(),globals())

        else:

            listax = [compare(self.n,index)]
        
        for k in listax: 
            
            if self.indices[k] == True:

                if self.name == 'Riemann' and k == 0: 

                    Riemann_list = construct('False',dim,4)

                    for p in iterprod(range(dim),repeat=self.n):

                        counta = p[0]
                        countb = p[1]
                        countc = p[2]
                        countd = p[3]

                        if Riemann_list[counta][countb][countc][countd] == False:
                            
                            string = 'self.%s[%s]'%(ten_call,k) 

                            for l in p:
                                    
                                string += '[%s]'%l

                            string = string + '=' + string + '.expand()'

                            exec(string,locals(),globals())

                            print('Fully Simplified:  ',counta,countb,countc,countd)

                            # Skew Symmetry

                            if Riemann_list[counta][countb][countd][countc] == False:

                                self.tensor[0][counta][countb][countd][countc] = -self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][counta][countb][countd][countc] = -self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[counta][countb][countd][countc] = True

                            if Riemann_list[countb][counta][countc][countd] == False:

                                self.tensor[0][countb][counta][countc][countd] = -self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][countb][counta][countc][countd] = -self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[countb][counta][countc][countd] = True

                            # Interchange Symmetry

                            if Riemann_list[countc][countd][counta][countb] == False:

                                self.tensor[0][countc][countd][counta][countb] = self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][countc][countd][counta][countb] = self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[countc][countd][counta][countb] = True

                            # Bianchi Identity (First)

                            if Riemann_list[counta][countc][countd][countb] == False and Riemann_list[counta][countd][countb][countc] == True:

                                self.tensor[0][counta][countc][countd][countb] = -self.tensor[0][counta][countb][countc][countd] - self.tensor[0][counta][countd][countb][countc]
                                self.tensor_sp[0][counta][countc][countd][countb] = -self.tensor_sp[0][counta][countb][countc][countd] - self.tensor_sp[0][counta][countd][countb][countc]

                                Riemann_list[counta][countc][countd][countb] = True

                            if Riemann_list[counta][countd][countb][countc] == False and Riemann_list[counta][countc][countd][countb] == True:

                                self.tensor[0][counta][countd][countb][countc] = -self.tensor[0][counta][countb][countc][countd] - self.tensor[0][counta][countc][countd][countb]
                                self.tensor_sp[0][counta][countd][countb][countc] = -self.tensor_sp[0][counta][countb][countc][countd] - self.tensor_sp[0][counta][countc][countd][countb]

                                Riemann_list[counta][countd][countb][countc] = True


                            Riemann_list[counta][countb][countc][countd] = True
                    
                else:
                
                    for p in iterprod(range(dim),repeat=self.n):
                                
                        string = 'self.%s[%s]'%(ten_call,k) 

                        for l in p:
                                
                            string += '[%s]'%l

                        if config.ord_status == False:

                            string = string + '= expand(' + string + ')'

                        else:

                            string = string + '= expand(tensor_series(' + string + '))'

                        #print(string)

                        exec(string,locals(),globals())

        if config.space_time == 1:

            self.space()

    def factor(self,index=None):


        """Factorizes each component of the tensor.

        Factorizes each component of the tensor. If index is given, only acts on the given indices combination.

        Parameters
        ----------
        index : str, optional
            string indicating the indices combination to apply the factorization (the default is None, it implies that it will factor all the inices combinations of the Tensor).

        Raises
        ------
        ValueError
            If the 'index' specified does not correspond to any possible indices combination.
        """

        dim = config.dim

        if (index not in self.sequence and index is not None) or index == '':

            raise ValueError('Bad index definition')

        
        if config.space_time == 1:

            dim = config.dim

            ten_call = 'tensor'

        else:

            dim = config.dim - 1

            ten_call = 'tensor_sp'
        

        if index is None:

            listax = list(range(2**self.n))

            if self.n == 0:


                self.tensor = eval("self.%s.factor()"%ten_call,locals(),globals())

        else:

            listax = [compare(self.n,index)]
        
        for k in listax: 
            
            if self.indices[k] == True:

                if self.name == 'Riemann' and k == 0: 

                    Riemann_list = construct('False',dim,4)

                    for p in iterprod(range(dim),repeat=self.n):

                        counta = p[0]
                        countb = p[1]
                        countc = p[2]
                        countd = p[3]

                        if Riemann_list[counta][countb][countc][countd] == False:
                            
                            string = 'self.%s[%s]'%(ten_call,k) 

                            for l in p:
                                    
                                string += '[%s]'%l

                            string = string + '=' + string + '.factor()'

                            exec(string,locals(),globals())

                            print('Fully Simplified:  ',counta,countb,countc,countd)

                            # Skew Symmetry

                            if Riemann_list[counta][countb][countd][countc] == False:

                                self.tensor[0][counta][countb][countd][countc] = -self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][counta][countb][countd][countc] = -self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[counta][countb][countd][countc] = True

                            if Riemann_list[countb][counta][countc][countd] == False:

                                self.tensor[0][countb][counta][countc][countd] = -self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][countb][counta][countc][countd] = -self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[countb][counta][countc][countd] = True

                            # Interchange Symmetry

                            if Riemann_list[countc][countd][counta][countb] == False:

                                self.tensor[0][countc][countd][counta][countb] = self.tensor[0][counta][countb][countc][countd]
                                self.tensor_sp[0][countc][countd][counta][countb] = self.tensor_sp[0][counta][countb][countc][countd]

                                Riemann_list[countc][countd][counta][countb] = True

                            # Bianchi Identity (First)

                            if Riemann_list[counta][countc][countd][countb] == False and Riemann_list[counta][countd][countb][countc] == True:

                                self.tensor[0][counta][countc][countd][countb] = -self.tensor[0][counta][countb][countc][countd] - self.tensor[0][counta][countd][countb][countc]
                                self.tensor_sp[0][counta][countc][countd][countb] = -self.tensor_sp[0][counta][countb][countc][countd] - self.tensor_sp[0][counta][countd][countb][countc]

                                Riemann_list[counta][countc][countd][countb] = True

                            if Riemann_list[counta][countd][countb][countc] == False and Riemann_list[counta][countc][countd][countb] == True:

                                self.tensor[0][counta][countd][countb][countc] = -self.tensor[0][counta][countb][countc][countd] - self.tensor[0][counta][countc][countd][countb]
                                self.tensor_sp[0][counta][countd][countb][countc] = -self.tensor_sp[0][counta][countb][countc][countd] - self.tensor_sp[0][counta][countc][countd][countb]

                                Riemann_list[counta][countd][countb][countc] = True


                            Riemann_list[counta][countb][countc][countd] = True
                    
                else:
                
                    for p in iterprod(range(dim),repeat=self.n):
                                
                        string = 'self.%s[%s]'%(ten_call,k) 

                        for l in p:
                                
                            string += '[%s]'%l

                        if config.ord_status == False:

                            string = string + '= factor(' + string + ')'

                        else:

                            string = string + '= factor(tensor_series(' + string + '))'

                        #print(string)

                        exec(string,locals(),globals())

        if config.space_time == 1:

            self.space()

    def display(self, index=None, aslist = None, simplify = False, spatial=None):

        """Displays the Tensor. 

        Displays the Tensor. By default it displays the tensor with all covariant indices unless 'index' is given.

        Parameters
        ----------
        index : str, optional
            string indicating the indices combination to display the Tensor (the default is None, it implies that the full covariant indices combination will be displayed).
        aslist : boolean, optional
            boolean indicating if the Tensor should be displayed in matrix form (False) or component by component as a list (True)
            (the default is None, it checks the rank of the Tensor, where if it is less than 2 it displays the tensor in matrix form, else it displays the Tensor as a list).
        simplify : boolean, optional
            if True, a simplification routin is applied to the displayed components. If False, the Tensor is displayed in its original form (the default is False).
        spatial : boolean, optional
            boolean indicating if only the spatial components of the Tensor should be displayed (True) 
            (the default value is None, it checks if the user is working only with space coordinates and if that's the case, only shows the spatial components, else, it shows the full Tensor).

        Raises
        ------
        ValueError
            If the 'index' specified does not correspond to any possible indices combination.

        Notes
        -----
        Note that if 'simplify' is True, the simplification is only performed in the display method and is not saved in the Tensor.

        """

       

        if config.space_time == False and spatial is None:

            spatial = True

        if spatial == True:

            return self.display_spatial(index, aslist, simplify)

        if core_calc == 'sp' and simplify == True:

            warn("The simplify argument is intended to be used only with giacpy.\n The result is not affected when using Sympy.")
        
        if index is None:

            index = self.sequence[0]

        if aslist == None:

            if self.n <= 2:

                aslist = False
            
            else:

                aslist = True

        dim = config.dim
        
        k = 0
        for i in self.sequence:
            if i == index:
                break 
            k += 1

        if core_calc == 'sp':

            init_printing()

        if k == len(self.sequence):

            raise ValueError('Bad index definition')

        if index == '' and self.n == 0: # Scalar

            display(self.tensor)
        
        
        elif aslist == False:

            # if k == len(self.sequence):

            #     raise ValueError('Bad index definition')

            if self.n == 1 and index == '^':
                
                if core_calc == 'sp':

                    display_IP(Array(self.tensor[k]).reshape(dim,1))

                elif core_calc == 'gp':

                    f = io.StringIO()

                    with redirect_stdout(f):

                        print(latex(giac(self.tensor[k]).transpose()))
                    out = f.getvalue()

                    out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                    display_IP(Math_IP(gp_pretty_latex(out)))

            else:

                if core_calc == 'sp':
            
                    display_IP(Array(self.tensor[k]))

                elif core_calc == 'gp':

                    if simplify == False:

                        f = io.StringIO()

                        if self.n != 1:
                            with redirect_stdout(f):
                                print(latex(matrix(self.tensor[k])))
                        else:
                             with redirect_stdout(f):
                                print(latex(giac(self.tensor[k])))
                        out = f.getvalue()

                        out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                        display_IP(Math_IP(gp_pretty_latex(out)))

                    else:

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print((self.tensor[k]))
                        out = f.getvalue()

                        string = sp_latex(sp_simplify(sp_Array(sp_sympify(out))))

                        display_IP(Math_IP(string))

            
        else:

            count = 0
                
            for p in iterprod(range(dim),repeat=self.n):
                    
                string = 'valor = self.tensor[%s]'%k
                    
                for l in p:
                        
                    string += '[%s]'%l

                exec(string,locals(),globals())
                
                if self.name in config.default_tensors:
                    
                    str_name = config.default_tensors[self.name]

                else:

                    str_name = self.name
                
                string = "{%s}"%str_name
                    
                i = 0
                    
                for l in index.split(','):
                        
                    if core_calc == 'gp':

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print(p[i])
                        out = f.getvalue()
                            
                        string = '%s{}%s{%s}'%(string,l,out)
                    
                    elif core_calc == 'sp':

                        string = '%s{}%s{%s}\,'%(string,l,str(p[i]))

                    i += 1
                    
                if valor != 0:
                        
                    #string += " = %s"% (latex(valor))
                    
                    #display_IP(Math(string))

                    if core_calc == 'gp':

                        if simplify == False:

                            display(valor,string)

                        else:

                            f = io.StringIO()
                            with redirect_stdout(f):
                                print(valor)
                            out = f.getvalue()

                            string2 = sp_latex(sp_simplify(sp_sympify(out)))

                            string = "%s = %s"%(string,string2)

                            display_IP(Math_IP(string))

                    elif core_calc == 'sp':

                        display(valor,string)

                    count += 1

            if count == 0:

                print('All components are zero')

    def display_spatial(self, index=None, aslist = None, simplify = False):

        if core_calc == 'sp' and simplify == True:

            warn("The simplify argument is intended to be used only with giacpy.\n The result is not affected when using Sympy.")
        
        if index is None:

            index = self.sequence[0]

        rank = self.n

        if aslist == None:

            if rank <= 2:

                aslist = False
            
            else:

                aslist = True

        dim = config.dim - 1 # Se elimina la dimension temporal
        
        k = 0
        for i in self.sequence:
            if i == index:
                break 
            k += 1

        if core_calc == 'sp':

            init_printing()

        if k == len(self.sequence):

            raise ValueError('Bad index definition')

        if index == '' and rank == 0: # Scalar

            display(self.tensor_sp)
        
        
        elif aslist == False:

            # if k == len(self.sequence):

            #     raise ValueError('Bad index definition')

            if rank == 1 and index == '^':
                
                if core_calc == 'sp':

                    display_IP(Array(self.tensor_sp[k]).reshape(dim,1))

                elif core_calc == 'gp':

                    f = io.StringIO()

                    with redirect_stdout(f):

                        print(latex(giac(self.tensor_sp[k]).transpose()))
                    out = f.getvalue()

                    out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                    display_IP(Math_IP(gp_pretty_latex(out)))

            else:

                if core_calc == 'sp':
            
                    display_IP(Array(self.tensor_sp[k]))

                elif core_calc == 'gp':

                    if simplify == False:

                        f = io.StringIO()

                        if rank != 1:
                            with redirect_stdout(f):
                                print(latex(matrix(self.tensor_sp[k])))
                        else:
                             with redirect_stdout(f):
                                print(latex(giac(self.tensor_sp[k])))
                        out = f.getvalue()

                        out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

                        display_IP(Math_IP(gp_pretty_latex(out)))

                    else:

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print((self.tensor_sp[k]))
                        out = f.getvalue()

                        string = sp_latex(sp_simplify(sp_Array(sp_sympify(out))))

                        display_IP(Math_IP(string))

            
        else:

            count = 0
                
            for p in iterprod(range(dim),repeat=rank):
                    
                string = 'valor = self.tensor_sp[%s]'%k

                for l in p:
                        
                    string += '[%s]'%l

                exec(string,locals(),globals())
                
                if self.name in config.default_tensors:
                    
                    str_name = config.default_tensors[self.name]

                else:

                    str_name = self.name
                
                string = "{%s}"%str_name
                    
                i = 0
                    
                for l in index.split(','):
                        
                    if core_calc == 'gp':

                        f = io.StringIO()
                        with redirect_stdout(f):
                            print(p[i])
                        out = f.getvalue()
                            
                        string = '%s{}%s{%s}'%(string,l,out)
                    
                    elif core_calc == 'sp':

                        string = '%s{}%s{%s}'%(string,l,str(p[i]))

                    i += 1
                    
                if valor != 0:
                        
                    #string += " = %s"% (latex(valor))
                    
                    #display_IP(Math(string))

                    if core_calc == 'gp':

                        if simplify == False:

                            display(valor,string)

                        else:

                            f = io.StringIO()
                            with redirect_stdout(f):
                                print(valor)
                            out = f.getvalue()

                            string2 = sp_latex(sp_simplify(sp_sympify(out)))

                            string = "%s = %s"%(string,string2)

                            display_IP(Math_IP(string))

                    elif core_calc == 'sp':

                        display(valor,string)

                    count += 1

            if count == 0:

                print('All components are zero')



def D(a,b):

    '''
    Derivative of a tensor "a" with respect to the index "b".

    a is an object of class Tdata
    b is a string that indicates the index, written as "_index"

    It returns an object of class Tdata. 

    -------------
    Example:

    A = D(G("^a, ^b"), "_c")

    We suggest to save this into a Tensor object as

    B = ten("B", 3)

    B.assign(A, "^a, ^b, _c")

    '''

    der_examine(b)

    if config.space_time == True:

        dim = config.dim

        coords = config.coords

    else:

        dim = config.dim - 1

        coords = config.coords_sp
    
    if isinstance(a,Tdata): #si es un Tdata
        
        old_rank = len(a.updn)
        
        if b[0] == '_':  # si es _

            #print('entro al _')

            if b[1:] in a.index_names:  #si se deriva con respecto a un indice repetido
                
                new_rank = old_rank - 1
            
                string_return = ''

                for i in range(len(a.index_names)):

                    if a.index_names[i] != b[1:]:

                        string_return += a.updn[i] + a.index_names[i] + ','

                string_return = string_return[:-1]


                temp = construct(0,dim,new_rank)

                new_index = ''
                
                k = 0

                while k < new_rank:

                    new_index +='[p[%d]]'%k

                    k += 1

                old_index = ''

                k = 0

                j = 0

                while k < old_rank:

                    if a.index_names[k] == b[1:]:

                        der_index ='q'

                        old_index +='[q]'

                    else:

                        old_index +='[p[%d]]'%j

                        j += 1
                    
                    k += 1
                
                k = 0

                exec(reload_all('config'),locals(),globals())

                for p in iterprod(range(dim),repeat = new_rank): 

                    var_temp = 0
                    
                    for q in range(dim):
    
                        string = 'diff(a.elements%s,coords[%s])'%(old_index,der_index)

                        var_temp += eval(string,locals(),globals())

                    string = 'temp%s = var_temp'%new_index

                    exec(string,locals(),globals())
                
                return Tdata(string_return, temp)
                
            else:                   #si se deriva con respecto a un indice no repetido 

                new_rank = old_rank + 1
            
                string_return = a.full_index +','+ b
                
                temp = construct(0,dim,new_rank)
                
                k = 0

                old_index = ''

                while k < old_rank:

                    old_index +='[p[%d]]'%k

                    k += 1
                
                k = 0
                
                new_index = ''
                
                while k < new_rank:

                    new_index +='[p[%d]]'%k

                    k += 1

                exec(reload_all('config'),locals(),globals())

                for p in iterprod(range(dim),repeat = new_rank): 
    
                    string = 'temp%s = diff(a.elements%s,%s)'%(new_index,old_index,coords[p[-1]])

                    exec(string,locals(),globals())
                
                return Tdata(string_return, temp)
            
        else:           # si es ^

            #print('entro al ^')

            new_rank = old_rank + 1
        
            new_var = 'dummy'

            while new_var in a.index_names:

                new_var += '0'

            new_var = '_' + new_var

            string_return = a.full_index +','+ new_var
            
            temp = construct(0,dim,new_rank)
            
            k = 0

            old_index = ''

            while k < old_rank:

                old_index +='[p[%d]]'%k

                k += 1
            
            k = 0
            
            new_index = ''
            
            while k < new_rank:

                new_index +='[p[%d]]'%k

                k += 1

            exec(reload_all('config'),locals(),globals())

            for p in iterprod(range(dim),repeat = new_rank): 

                string = 'temp%s = diff(a.elements%s,%s)'%(new_index,old_index,coords[p[-1]])

                exec(string,locals(),globals())

            new_var2 = '^' + new_var[1:]

            string_return = a.full_index +','+ new_var

            if b[1:] not in a.index_names:

                g_string = b + ',' + new_var2

                #print(g_string, string_return)

                return config.g(g_string)*Tdata(string_return, temp)

            else:

                new_var = 'dummy'

                while new_var in a.index_names or new_var == new_var2[1:]:

                    new_var += '0'

                new_var = '^' + new_var

                g_string = new_var + ',' + new_var2

                #print(g_string, string_return)

                temp_return = config.g(g_string)*Tdata(string_return, temp)

                temp_return.full_list[0] = '^' + temp_return.full_list[-1][1:]

                string = ','.join(temp_return.full_list)     

                temp_return = Tdata(string,temp_return.elements)

                return temp_return.auto_sum()

    else: # si var es un int o sympy o symengine

        temp = construct(0,dim,1)

        string_return = b

        exec(reload_all('config'),locals(),globals())

        for p in range(dim):

            string = 'temp[%d] = diff(a,%s)'%(p,coords[p])

            exec(string,locals(),globals())
        
        return Tdata(string_return, temp)

def C(a,b):

    '''
    Covariant derivative of a tensor "a" with respect to the index "b".

    a is an object of class Tdata
    b is a string that indicates the index, written as "_index"

    It returns an object of class Tdata. 

    -------------
    Example:

    A = C(G("^a, ^b"), "_c")

    We suggest to save this into a Tensor object as

    B = ten("B", 3)

    B.assign(A, "^a, ^b, _c")

    '''

    der_examine(b)

    if config.space_time == True:

        dim = config.dim

    else:

        dim = config.dim - 1

    if isinstance(a,Tdata): #si es un Tdata

        if config.christ is None:

            print("Christofel Symbols not calculated.")

            raise(NotImplementedError)

        was_up = False
        was_rep = False

        if b[0] == '^':  # si es _

            was_up = True

            b = '_' + b[1:]

        if b[1:] in a.index_names:  #si se deriva con respecto a un indice repetido

            rep_symbol = b[1:][:]

            new_symbol = 'dummy'
            i = 0

            was_symbol = b[1:][:]

            was_rep = True

            while new_symbol in a.index_names or new_symbol == b[1:]:

                new_symbol += '%d'%i
                i += 1

            b = b[0] + new_symbol[:]

        old_rank = len(a.updn)

        new_rank = old_rank + 1
    
        string_return = a.full_index +','+ b  # _i,_j + _k

        temp_full_index = '%s'%a.full_index
        
        return_tensor = construct(0,dim,new_rank)

        a_list_index = a.full_index.split(',')

        k = 0

        old_index = ''

        while k < old_rank:

            old_index +='[p[%d]]'%k

            k += 1
        
        k = 0
        
        new_index = ''
        
        while k < new_rank:

            new_index +='[p[%d]]'%k

            k += 1

        dummy = 'dummy'

        i = 0

        while dummy in a.index_names or dummy == b[1:]:

            dummy += '%d'%i
            i += 1

        exec(reload_all('config'),locals(),globals())
            
        var = D(a,b)

        for num, ind in enumerate(a_list_index):  # ['^x','_y']

            tdata_str = ''

            if ind[0] == '^':
        
                chrstr = '^%s,_%s,_%s'%(ind[1:],dummy,b[1:])

                for l in range(len(a_list_index)):

                    if l == num:

                        tdata_str += '^%s,'%dummy
                    else:
                        tdata_str += '%s,'%a_list_index[l]

                tdata_str = tdata_str[:-1]

                TEMP_tdata = Tdata(tdata_str,a.elements)

                var = var + config.christ(chrstr)*TEMP_tdata

            elif ind[0] == '_':
        
                chrstr = '^%s,_%s,_%s'%(dummy,ind[1:],b[1:])

                for l in range(len(a_list_index)):

                    if l == num:

                        tdata_str += '_%s,'%dummy
                    else:
                        tdata_str += '%s,'%a_list_index[l]

                tdata_str = tdata_str[:-1]

                TEMP_tdata = Tdata(tdata_str,a.elements)

                var = var - config.christ(chrstr)*TEMP_tdata

        # si el indice c/r al que se deriva estaba arriba entonces multiplicamos por la metrica inversa

        # hay que revisar si esta dando bien porque hay q constatar el orden de los indices

        # esta hecho CON LA IDEA DE QUE, HAY Q REVISARLO...., LA IDEA ES: QUE SIEMPRE EL INDICE QUE SE HEREDA QUEDA AL FINAL
        # POR ORDEN, ASI QUE SIEMPRE HAY QUE REEMPLAZAR SOLO EL ULTIMO INDICE

        if was_up == True:  # solo sube el indice NO SE FIJA SI HAY REPETICION

            dummy = 'dummy'

            i = 0

            while dummy in var.index_names:

                dummy += '%d'%i
                i += 1

            moved_symbol = var.index_names[-1][:]
            
            str_temp = var.full_index[:]

            var = var * config.g('^%s,^%s'%(dummy,moved_symbol))

            str_temp = var.full_index[:].replace('dummy', moved_symbol)

            var = Tdata(str_temp,var.elements)

        if was_rep == True:

            moved_symbol = var.index_names[-1][:]

            str_temp = ''

            for i,name in enumerate(var.index_names):

                if i == len(var.index_names) - 1:

                    str_temp += '^' + var.updn[i][1:] + rep_symbol +','

                else: 
                    
                    str_temp += var.updn[i] + var.index_names[i]  +','

            str_temp = str_temp[:-1]

            var = Tdata(str_temp,var.elements)     
######################### SON MUCHOS CASOS ARRIBA REPETIDOS ABAJO REPETIDOS, ABAJO SIN REPETIR ARRIBA SIN REPETIR

        #print(var.full_index)

        #print('antes d entrar al autosum')

        return var.auto_sum()
        
    else: # si var es un int o sympy o symengine
    
        return D(a,b)

def create(name,TD):

    '''
    Ponerla como function de un TDATA entonces hacemos algo como A = TD.save('name')
    '''
    
    rank = len(TD.full_list)
    
    T = new_ten(name,rank)
    
    T.assign(TD,TD.full_index)
    
    T.factor((',').join(TD.updn))
    
    return T

def determinant(T):

    if not isinstance(T,Tdata):

        raise(ValueError("T must ve a Tdata instance."))
    
    if len(T.updn) != 2: 
        
        raise(ValueError("Must be a rank 2 tensor"))
        
    if core_calc == 'sp':
        
        return det(Matrix(T.elements))
        
    elif core_calc == 'gp':
        
        raise(NotImplementedError("Method not implemented for giacpy"))
        
        

def create_tensor(T_name,n):
    
    create_temp(name,obj)

    T_object = Tensor(T_name,n)

    return config.create_ten(T_name,T_object)

def tensor_series(element):

    '''
    Compute the series of an element.
    '''

    if core_calc == 'gp':

        result = series(element,config.ord_var,0,config.ord_n)

    elif core_calc == 'sp':

        result = series(expand(element), config.ord_var,0, config.ord_n+1)

    return result


import tqdm
from itertools import product as iterprod 
from pytearcat.tensor.misc import new_ten,reload_all
from pytearcat.tensor.core import config
from pytearcat.tensor.tensor import construct, tensor_series, Tensor
from .riemann import calculate_riemann
from .ricci import calculate_ricci, calculate_ricci_scalar
from pytearcat.tensor.core.core import core_calc,display_IP,Latex_IP

if core_calc == 'sp':

    from pytearcat.tensor.core.core import diff, Rational, sympify, factor

elif core_calc == 'gp':

    from pytearcat.tensor.core.core import diff, divide, simplify

def calculate_einstein(All = False):

    global Einstein

    dim = config.dim

    g = config.g.tensor

    if config.G is None:

        Einstein =config.create_ten('Einstein',Tensor('Einstein',2))

        config.G = Einstein


    else:

        Einstein = config.G

    if config.ricci is None:

        Ricci_local = calculate_ricci()

    else: 

        Ricci_local = config.ricci
    
    if config.ricciS is None:

        R = calculate_ricci_scalar()
    
    else:

        R = config.ricciS

        
    #---

    if Einstein.indices[0] == False:

        Einstein_list = construct('False',dim,2)

        display_IP(Latex_IP(r'Einstein Tensor $G_{\alpha \beta}$'))

        for p in tqdm.tqdm_notebook(iterprod(range(config.dim),repeat=2),total=config.dim**2):

            m = p[0]
            n = p[1]

            if Einstein_list[m][n] == False:

                if core_calc == 'sp':

                    EinstTemp = Ricci_local.tensor[0][m][n] - Rational(1,2)*(g[0][m][n])*R

                elif core_calc == 'gp':

                    EinstTemp = Ricci_local.tensor[0][m][n] - divide(1,2)[0]*(g[0][m][n])*R

                if config.ord_status == True:

                    Einstein.tensor[0][m][n] = tensor_series(EinstTemp)

                else:

                    if core_calc == 'sp':

                        Einstein.tensor[0][m][n] = sympify(factor(EinstTemp))

                    elif core_calc == 'gp':

                        Einstein.tensor[0][m][n] = simplify(EinstTemp)

                Einstein_list[m][n] = True

                # Simmetry

                Einstein.tensor[0][n][m] = Einstein.tensor[0][m][n]

                Einstein_list[n][m] = True

        Einstein.indices[0] = True

    else:

        display_IP(Latex_IP(r"Einstein Tensor $G_{\alpha \beta}$ already calculated"))

    if All == True:

        Einstein.complete('_,_')

    Einstein.space()

    return Einstein


import giacpy as __gp
from giacpy import latex
import io
from re import findall, search
from IPython.display import display as display_IP, Math as Math_IP
from contextlib import redirect_stdout
from .tdata import Tdata as _Tdata
from .ptseries import ord_status as ord_status,ord_var,ord_n
#from pytearcat.tensor.core import ptseries
from pytearcat.tensor.core import greek
from pytearcat.tensor.core import fun

def reload_all(new_module):

    ''' 

    This function reloads all the variables in __all__ from module
        
    Thanks to hurturk on https://stackoverflow.com/questions/44492803/python-dynamic-import-how-to-import-from-module-name-from-variable?newreg=9312458d429647fc8ec2a72e5655d197
        
    '''

    string = "globals().update({{n: getattr(module, n) for n in module.__all__}} if hasattr(module, '_all_') else {k: v for (k, v) in module.__dict__.items() if not k.startswith('_')})"

    string = string.replace('module',new_module)

    return string

def get_name(element):
        
    f = io.StringIO()
    with redirect_stdout(f):
        print(element)
    element = f.getvalue()

    return element[:-1]

def tolatex(element):

    f = io.StringIO()
    with redirect_stdout(f):
        print(latex(element))
    element = f.getvalue()

    string =  element[1:-2]

    return string

def gp_pretty_order(element):
    
    #ord_n = ord_n

    global ord_status,ord_var,ord_n

    ord_var = get_name(ord_var)
    
    if ord_var in greek.greek_dict:

        ord_var = r"\%s"%ord_var

    structure = r"\%s\^\{(\d+)\} \\operatorname\{\\mathrm\{order\\_size\} \}\\left\(\%s\\right\)"%(ord_var,ord_var)

    exponent = findall(structure,element)

    if len(exponent) > 0:

        structure = r"%s^{%s} \operatorname{\mathrm{order\_size} }\left(%s\right)"%(ord_var,exponent[0],ord_var)

        final = element.replace(r"%s"%structure,r"\mathrm{O}\left(%s^{%s}\right)"%(ord_var,exponent[0]))
        
    elif len(exponent) == 0:

        structure = r"%s \operatorname{\mathrm{order\_size} }\left(%s\right)"%(ord_var,ord_var)

        final = element.replace(r"%s"%structure,r"\mathrm{O}\left(%s\right)"%ord_var)

    return final

def gp_pretty_latex(element):
    
    '''
    It takes a giacpy latex expression and returns a latex string that is similar to the sympy notation.
    
        - It rewrites the derivatives
        - It rewrites the expansion order
    
    '''

    global ord_status,ord_var,ord_n 
    names = {}
    variables = []

    result = element[:]
    
    greek_dict =greek.greek_dict

    for i in fun.fun:

        x = r'%s'%str(latex(i))[1:].split("\left")[0].replace('\\','').replace('{','\\{').replace('}','\\}').replace('mathrm','\\\\mathrm')

        names[x] = str(i).split('(')[1][:-1]

    for i in names:

        j = findall(r'(?<=\W)%s\^\{\\left\((.*?)\\right\)\}'%i,element)
        
        for k in j:

            ind = k.split(',')

            ini,fin = search(r'%s\^\{\\left\(%s\\right\)\}'%(i,k),element).span()
            
            while element[ini] != ' ' and element[ini] != ',' and element[ini] != '\\' and element[ini] != '-' and element[ini] != '+':

                ini -= 1

            string = ''

            l_dict = {}

            for l in ind: 

                l_dict[l] = ind.count(l)  
                
            for l in l_dict:  # L = '1'

                y = names[i].split(',')[int(l)-1]
                    
                if y in greek_dict.keys():
                        
                    y = r'\\%s'%y
                
                if l_dict[l] == 1:
                    
                    string += r'\frac{\partial }{\partial %s}'%y

                else:

                    string += r'\frac{\partial^%d }{\partial %s^%d}'%(l_dict[l],y,l_dict[l])

            string+= i

            result = result.replace(element[ini:fin],string)    

    for func in names:

        if func in greek_dict.keys():

            result = result.replace(func,greek_dict[func])
            
    if ord_status == True:

        result = gp_pretty_order(result)

    return result.replace('\\\\','\\').replace("\"","").replace("\\{","{").replace("\\}","}")

def reload_series():

    global ord_status,ord_var,ord_n

    from .ptseries import ord_status,ord_var,ord_n

    return ord_status,ord_var,ord_n


class gpcore(__gp.giacpy.Pygen):
 
    def __init__(self,x):
        
        super().__init__()

    def __getitem__(self,rollno):

        return gpcore(super().__getitem__(rollno))

    def __setitem__(self,rollno,name):

        return gpcore(super().__setitem__(rollno,name))

    def __str__(self):

        #print("STR")

        return super().__str__()

    def _repr_html_(self):

        #print("llamando a html")


        ord_status,ord_var,ord_n = reload_series()


        if ord_status == True:

           a = expand(self)

        else:

            a = self

        string = tolatex(a)

        string = gp_pretty_latex(string)

        #display_IP(Math_IP(string))

        return "$$ %s $$"%string

    def __repr__(self):    

        #print("llamando a repr")

        

        return ""
        #return "llamando a repr" #str(self)
        
    def __mul__(self,other):
        
        if isinstance(other,_Tdata):
        
            return other*self
        
        else:

            return gpcore(super().__mul__(other))  
        
    def __rmul__(self,other):
        
        if isinstance(other,_Tdata):
        
            return self*other
        
        else:
            
            return gpcore(super().__rmul__(other))  
        
    def __add__(self,other):
        
        return gpcore(super().__add__(other))  

    def __radd__(self,other):
        
        return gpcore(super().__radd__(other))  
    
    def __sub__(self,other):
        
        return gpcore(super().__sub__(other))  
    
    def __rsub__(self,other):
        
        return gpcore(super().__rsub__(other))  
    
    def __truediv__(self,other):
        
        return gpcore(super().__truediv__(other))  
    
    def __rtruediv__(self,other):
        
        return gpcore(super().__rtruediv__(other))  
    
    def __pow__(self,other):
        
        return gpcore(super().__pow__(other)) 
    
    def __rpow__(self,other):
        
        return gpcore(super().__rpow__(other))  

    def __neg__(self):
        
        return gpcore(super().__neg__())  

    def __pos__(self):
        
        return gpcore(super().__pos__())  
    
    def simplify(self):
        
        return  gpcore(super().simplify())#simplify(self)

    def expand(self):

        return gpcore(super().expand())

    def factor(self):

        return gpcore(super().factor())

    

#trigonometric functions

def sin(x):

    return gpcore(__gp.sin(x))

def cos(x):

    return gpcore(__gp.cos(x))

def tan(x):

    return gpcore(__gp.tan(x))

def csc(x):

    return gpcore(__gp.csc(x))

def sec(x):

    return gpcore(__gp.sec(x))

def cot(x):

    return gpcore(__gp.cot(x))

# arc trigonometric functions

def asin(x):

    return gpcore(__gp.asin(x))

def acos(x):

    return gpcore(__gp.acos(x))

def atan(x):

    return gpcore(__gp.atan(x))

def acsc(x):

    return gpcore(__gp.acsc(x))

def asec(x):

    return gpcore(__gp.asec(x))

def acot(x):

    return gpcore(__gp.acot(x))

# hyperbolic functions

def sinh(x):

    return gpcore(__gp.sinh(x))

def cosh(x):

    return gpcore(__gp.cosh(x))

def tanh(x):

    return gpcore(__gp.tanh(x))

def csch(x):

    return gpcore(__gp.csch(x))

def sech(x):

    return gpcore(__gp.sech(x))

def coth(x):

    return gpcore(__gp.coth(x))

# arc hyperbolic functions

def asinh(x):

    return gpcore(__gp.asinh(x))

def acosh(x):

    return gpcore(__gp.acosh(x))

def atanh(x):

    return gpcore(__gp.atanh(x))

def acsch(x):

    return gpcore(__gp.acsch(x))

def asech(x):

    return gpcore(__gp.asech(x))

def acoth(x):

    return gpcore(__gp.acoth(x))

# exp

def exp(x):

    return gpcore(__gp.exp(x))

# log

def log(x):

    return gpcore(__gp.log(x))

def log10(x):

    return gpcore(__gp.log10(x))

def logb(x,b):

    return gpcore(__gp.logb(x,b))

# sqrt

def sqrt(x):

    return gpcore(__gp.sqrt(x))

# abs

def abs(x):

    return gpcore(__gp.abs(x))

# general functions

def simplify(x):

    return gpcore(__gp.simplify(x))

def expand(x):

    return gpcore(__gp.expand(x))

def factor(x):

    return gpcore(__gp.factor(x))

def series(f,epsilon,x0,n):

    return gpcore(__gp.series(f,epsilon,x0,n))

def giac(x):

    return gpcore(__gp.giac(x))

def expand(x):

    return gpcore(__gp.expand(x))

def diff(f,x,n = 1):

    return gpcore(__gp.diff(f,x,n))

def divide(a,b):

    return gpcore(__gp.divide(a,b))

def factor(x):

    return gpcore(__gp.factor(x))

def zeros(a,b):

    return gpcore(__gp.zeros(a,b))

def coeff(a,x,n):

    return gpcore(__gp.coeff(a,x,n))

def idn(n):

    return gpcore(__gp.idn(n))

def matrix(x):

    return gpcore(__gp.matrix(x))


# Greek Alphabet

greek = [r'\alpha',  r'\beta', r'\gamma', r'\delta',r'\mu', r'\nu',r'\varrho', r'\rho',r'\sigma',r'\tau', r'\omega', r'\kappa', r'\lambda',r'\psi', r'\eta',r'\vartheta',  r'\theta', r'\iota',r'\zeta', r'\omicron', r'\pi', r'\upsilon', r'\varphi',r'\phi',r'\chi']

greek_dict = {'alpha' : r'\alpha', 'beta' : r'\beta', 'gamma' : r'\gamma', 'delta': r'\delta', 'mu' : r'\mu', 'nu' : r'\nu','varrho' : r'\rho', 'rho' : r'\rho', 'sigma' : r'\sigma', 'tau' : r'\tau', 'omega' : r'\omega', 'kappa' : r'\kappa', 'lambda' : r'\lambda','psi': r'\psi', 'eta' : r'\eta','vartheta' : r'\theta', 'theta' : r'\theta', 'iota': r'\iota', 'zeta' : r'\zeta', 'omicron' : r'\omicron', 'pi' : r'\pi', 'upsilon' : r'\upsilon','varphi' : r'\phi', 'phi': r'\phi', 'chi': r'\chi',\
    'Alpha' : r'\Alpha', 'Beta' : r'\Beta', 'Gamma' : r'\Gamma', 'Delta': r'\Delta', 'Mu' : r'\Mu', 'Nu' : r'\Nu', 'Rho' : r'\Rho', 'Sigma' : r'\Sigma', 'Tau' : r'\Tau', 'Omega' : r'\Omega', 'Kappa' : r'\Kappa', 'Lambda' : r'\Lambda', 'Eta' : r'\Eta', 'Theta' : r'\Theta', 'Iota': r'\Iota', 'Zeta' : r'\Zeta', 'Omicron' : r'\Omicron', 'Pi' : r'\Pi', 'Upsilon' : r'\Upsilon', 'Phi' : r'\Phi', 'Chi' : r'\Chi', 'Psi' : r'\Psi'}



from .greek import *

from .core import core_calc
from .fun import fun
from .ptseries import *

if core_calc == 'sp':

    from .core import Symbol, Function

elif core_calc == 'gp':

    from .core import giac

# Quiero que las cosas se ejecuten en misc y pasarle el elemento listo a config para que lo guarde en las listas correspondientes
# la razon de esto es que si queremos guardar tensores, habría que hacer import de la clase tensor y tensorclass ya importa a config

def create_var(i):

    if core_calc == 'sp':

        def_var = "%s = Symbol('%s', real=True)"%(i,i)

    elif core_calc == 'gp':

        def_var = "%s = giac('%s')"%(i,i)

    exec(def_var,locals(),globals())
    
    append_str = "__all__.append('%s')"%i
    exec(append_str,globals())

    append_str = 'var.append(%s)'%i
    exec(append_str,globals())
    
    return var[-1]

def create_con(i):

    if core_calc == 'gp':

        string = "%s = giac('%s')"%(i,i)

    elif core_calc == 'sp':

        string = "%s = Symbol('%s', real=True, constant = True)"%(i,i)

    exec(string,locals(),globals())
    
    string = "__all__.append('%s')"%i
    exec(string,globals())

    string = 'con.append(%s)'%i
    exec(string,globals())

    return con[-1]

def create_fun(f_symbol,var_symbol):

    if core_calc == 'gp':

        str5 = "%s=giac('%s(%s)')"%(f_symbol,f_symbol,var_symbol)

    elif core_calc == 'sp':
    
        str5 = "%s=Function('%s', real=True, nonzero=True)(%s)"%(f_symbol,f_symbol,var_symbol)

    exec(str5,locals(),globals())  

    assign = "__all__.append('%s')"%str(f_symbol)

    exec(assign,globals())

    assign = 'fun.append(%s)'%str(f_symbol)

    exec(assign,globals())

    return fun[-1]


def create_default(T_name,T_obj):

    if T_name == 'Christoffel':

        christ = T_obj

        ten.append(christ)

    elif T_name == 'Riemann':

        riemann = T_obj

        ten.append(riemann)

    elif T_name == 'Ricci':

        ricci = T_obj

        ten.append(ricci)

    elif T_name == 'RicciS':

        ricciS = T_obj

        ten.append(ricciS)

    elif T_name == 'Einstein':

        G = T_obj

        ten.append(G)




def create_ten(T_name,T_obj):

    if T_name in default_tensors:

        create_default(T_name,T_obj)

    string = "%s = T_obj"%T_name
    exec(string,locals(),globals())

    string = "__all__.append('%s')"%str(T_name)
    exec(string,globals())

    string = 'ten.append(%s)'%T_name
    exec(string,globals())

    #Este return es para que retorne el objeto que quedo definido aca en config.

    return ten[-1]

def create_temp(name,obj):

    string = "%s = obj"%name
    exec(string,locals(),globals())

    string = "__all__.append('%s')"%str(name)
    exec(string,globals())

    string = "temp.append('%s')"%str(name)
    exec(string,globals())

    string = "temp_name.append('%s')"%str(name)
    exec(string,globals())

    #Este return es para que retorne el objeto que quedo definido aca en config.

    return temp[-1]

# Manifold dimension
dim = 0

# Defined Metric

g_status = False

ds = ''

coords = {}

coords_sp = {}

geo = None

geocoords = {}

space_time = True

var = [] # Lista con variables de tipo sympy

con = [] # Lista con constantes de tipo sympy

#fun = [] # Lista con funciones de tipo sympy

# Tensores tipicos -------- # Proteger estos nombres para que no puedan ser sobreescritos. Que newten arroje error 

g = None  # Metrica

christ = None  # Christoffel

riemann = None # Riemann

ricci = None # Ricci Tensor

ricciS = None # Ricci Scalar

G = None # Einstein Tensor

LeviCivita_Symbol = None # LeviCivita Symbol

LeviCivita_Tensor = None # LeviCivita Tensor

default_tensors = {'Christoffel' : '\Gamma' , 'Einstein' : 'G', 'Riemann': 'R', 'Ricci' : 'R', 'RicciS' : 'R','g':'g','G' :'G'}

# ----------------------------

ten = [] # Lista de objetos de clase Tensor

temp = []

temp_name = []

# Order  (epsilon)

#ord_status = False

#ord_var = ''

#ord_n = 0

__all__ = ['g','christ','riemann','ricci','ricciS','G','space_time']